#############################################################################
##
## Copyright (C) 2021 Riverbank Computing Limited.
## Copyright (C) 2014 Digia Plc
## All rights reserved.
## For any questions to Digia, please use contact form at http://qt.digia.com
##
## This file is part of the QtDataVisualization module.
##
## Licensees holding valid Qt Enterprise licenses may use this file in
## accordance with the Qt Enterprise License Agreement provided with the
## Software or, alternatively, in accordance with the terms contained in
## a written agreement between you and Digia.
##
## If you have questions regarding the use of this file, please use
## contact form at http://qt.digia.com
##
#############################################################################


from PyQt6.QtCore import QObject, QPoint, QSize, QSizeF, Qt, QTimer
from PyQt6.QtDataVisualization import (Q3DBars, Q3DCamera, Q3DTheme,
        QAbstract3DGraph, QAbstract3DSeries, QBar3DSeries,
        QItemModelBarDataProxy)
from PyQt6.QtGui import QFont
from PyQt6.QtWidgets import (QApplication, QSizePolicy, QTableWidget,
        QVBoxLayout, QWidget)


class GraphDataGenerator(QObject):

    def __init__(self, bargraph, tableWidget):
        super().__init__()

        self.m_graph = bargraph
        self.m_tableWidget = tableWidget

        self.m_graph.setBarThickness(1.0)
        self.m_graph.setBarSpacing(QSizeF(0.2, 0.2))

        self.m_graph.setSelectionMode(
                QAbstract3DGraph.SelectionFlag.SelectionItemAndRow |
                QAbstract3DGraph.SelectionFlag.SelectionSlice)

        self.m_graph.activeTheme().setType(Q3DTheme.Theme.ThemeDigia)
        self.m_graph.activeTheme().setFont(QFont('Impact', 20))

        self.m_graph.scene().activeCamera().setCameraPreset(
                Q3DCamera.CameraPreset.CameraPresetFront)

    def start(self):
        self.setupModel()

        # The table needs to be shown before the size of its headers can be
        # accurately obtained so we postpone it a bit.
        QTimer.singleShot(100, self.fixTableSize)

    def setupModel(self):
        days = ("Monday", "Tuesday", "Wednesday", "Thursday", "Friday",
            "Saturday", "Sunday")

        weeks = ("week 1", "week 2", "week 3", "week 4", "week 5")

        #         Mon  Tue  Wed  Thu  Fri  Sat  Sun
        hours = ((2.0, 1.0, 3.0, 0.2, 1.0, 5.0, 10.0),  # week 1
                 (0.5, 1.0, 3.0, 1.0, 2.0, 2.0, 3.0),   # week 2
                 (1.0, 1.0, 2.0, 1.0, 4.0, 4.0, 4.0),   # week 3
                 (0.0, 1.0, 0.0, 0.0, 2.0, 2.0, 0.3),   # week 4
                 (3.0, 3.0, 6.0, 2.0, 2.0, 1.0, 1.0))   # week 5

        self.m_graph.rowAxis().setTitle("Week of year")
        self.m_graph.columnAxis().setTitle("Day of week")
        self.m_graph.valueAxis().setTitle("Hours spent on the Internet")
        self.m_graph.valueAxis().setLabelFormat("%.1f h")

        self.m_tableWidget.setRowCount(len(weeks))
        self.m_tableWidget.setColumnCount(len(days))
        self.m_tableWidget.setHorizontalHeaderLabels(days)
        self.m_tableWidget.setVerticalHeaderLabels(weeks)
        self.m_tableWidget.setHorizontalScrollBarPolicy(
                Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.m_tableWidget.setVerticalScrollBarPolicy(
                Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.m_tableWidget.setCurrentCell(-1, -1)

        model = self.m_tableWidget.model()

        for week in range(len(weeks)):
            for day in range(len(days)):
                index = model.index(week, day)
                model.setData(index, hours[week][day])

    def selectFromTable(self, selection):
        self.m_tableWidget.setFocus()
        self.m_tableWidget.setCurrentCell(selection.x(), selection.y())

    def selectedFromTable(self, currentRow, currentColumn, previousRow,
            previousColumn):
        self.m_graph.seriesList()[0].setSelectedBar(
                QPoint(currentRow, currentColumn))

    def fixTableSize(self):
        width = self.m_tableWidget.horizontalHeader().length()
        width += self.m_tableWidget.verticalHeader().width()
        self.m_tableWidget.setFixedWidth(width + 2)

        height = self.m_tableWidget.verticalHeader().length()
        height += self.m_tableWidget.horizontalHeader().height()
        self.m_tableWidget.setFixedHeight(height + 2)


if __name__ == '__main__':

    import sys

    app = QApplication(sys.argv)
    graph = Q3DBars()
    container = QWidget.createWindowContainer(graph)

    screenSize = graph.screen().size()
    container.setMinimumSize(
            QSize(int(screenSize.width() / 2), int(screenSize.height() / 2)))
    container.setMaximumSize(screenSize)
    container.setSizePolicy(QSizePolicy.Policy.Expanding,
            QSizePolicy.Policy.Expanding)
    container.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

    widget = QWidget()
    layout = QVBoxLayout(widget)
    tableWidget = QTableWidget()
    layout.addWidget(container, 1)
    layout.addWidget(tableWidget, 1, Qt.AlignmentFlag.AlignHCenter)

    tableWidget.setSizePolicy(QSizePolicy.Policy.Expanding,
            QSizePolicy.Policy.Fixed)
    tableWidget.setAlternatingRowColors(True)

    widget.setWindowTitle("Hours spent on the Internet")

    # Since we are dealing with QTableWidget, the model will already have data
    # sorted properly into rows and columns, so we simply set the
    # useModelCategories property to True to utilise this.
    proxy = QItemModelBarDataProxy(tableWidget.model())
    proxy.setUseModelCategories(True)

    series = QBar3DSeries(proxy)
    series.setMesh(QAbstract3DSeries.Mesh.MeshPyramid)
    graph.addSeries(series)

    generator = GraphDataGenerator(graph, tableWidget)
    series.selectedBarChanged.connect(generator.selectFromTable)
    tableWidget.currentCellChanged.connect(generator.selectedFromTable)

    widget.show()
    generator.start()
    sys.exit(app.exec())



#############################################################################
##
## Copyright (C) 2021 Riverbank Computing Limited.
## Copyright (C) 2014 Digia Plc
## All rights reserved.
## For any questions to Digia, please use contact form at http://qt.digia.com
##
## This file is part of the QtDataVisualization module.
##
## Licensees holding valid Qt Enterprise licenses may use this file in
## accordance with the Qt Enterprise License Agreement provided with the
## Software or, alternatively, in accordance with the terms contained in
## a written agreement between you and Digia.
##
## If you have questions regarding the use of this file, please use
## contact form at http://qt.digia.com
##
#############################################################################


from PyQt6.QtCore import pyqtSignal, QObject, QSize, Qt
from PyQt6.QtDataVisualization import (Q3DBars, Q3DCamera, Q3DTheme,
        QAbstract3DGraph, QAbstract3DSeries, QBar3DSeries, QBarDataItem,
        QCategory3DAxis, QValue3DAxis)
from PyQt6.QtGui import QFont
from PyQt6.QtWidgets import (QApplication, QCheckBox, QComboBox, QFontComboBox,
        QHBoxLayout, QLabel, QPushButton, QSizePolicy, QSlider, QVBoxLayout,
        QWidget)


class GraphModifier(QObject):

    tempOulu = (
        (-6.7, -11.7, -9.7, 3.3, 9.2, 14.0, 16.3, 17.8, 10.2, 2.1, -2.6, -0.3),
        (-6.8, -13.3, 0.2, 1.5, 7.9, 13.4, 16.1, 15.5, 8.2, 5.4, -2.6, -0.8),
        (-4.2, -4.0, -4.6, 1.9, 7.3, 12.5, 15.0, 12.8, 7.6, 5.1, -0.9, -1.3),
        (-7.8, -8.8, -4.2, 0.7, 9.3, 13.2, 15.8, 15.5, 11.2, 0.6, 0.7, -8.4),
        (-14.4, -12.1, -7.0, 2.3, 11.0, 12.6, 18.8, 13.8, 9.4, 3.9, -5.6, -13.0),
        (-9.0, -15.2, -3.8, 2.6, 8.3, 15.9, 18.6, 14.9, 11.1, 5.3, 1.8, -0.2),
        (-8.7, -11.3, -2.3, 0.4, 7.5, 12.2, 16.4, 14.1, 9.2, 3.1, 0.3, -12.1),
        (-7.9, -5.3, -9.1, 0.8, 11.6, 16.6, 15.9, 15.5, 11.2, 4.0, 0.1, -1.9)
    )

    tempHelsinki = (
        (-3.7, -7.8, -5.4, 3.4, 10.7, 15.4, 18.6, 18.7, 14.3, 8.5, 2.9, 4.1),
        (-1.2, -7.5, 3.1, 5.5, 10.3, 15.9, 17.4, 17.9, 11.2, 7.3, 1.1, 0.5),
        (-0.6, 1.2, 0.2, 6.3, 10.2, 13.8, 18.1, 15.1, 10.1, 9.4, 2.5, 0.4),
        (-2.9, -3.5, -0.9, 4.7, 10.9, 14.0, 17.4, 16.8, 13.2, 4.1, 2.6, -2.3),
        (-10.2, -8.0, -1.9, 6.6, 11.3, 14.5, 21.0, 18.8, 12.6, 6.1, -0.5, -7.3),
        (-4.4, -9.1, -2.0, 5.5, 9.9, 15.6, 20.8, 17.8, 13.4, 8.9, 3.6, 1.5),
        (-3.5, -3.2, -0.7, 4.0, 11.1, 13.4, 17.3, 15.8, 13.1, 6.4, 4.1, -5.1),
        (-4.8, -1.8, -5.0, 2.9, 12.8, 17.2, 18.0, 17.1, 12.5, 7.5, 4.5, 2.3)
    )

    months = (
        "January", "February", "March", "April", "May", "June",
        "July", "August", "September", "October", "November", "December"
    )

    years = ("2006", "2007", "2008", "2009", "2010", "2011", "2012", "2013")

    shadowQualityChanged = pyqtSignal(int)
    backgroundEnabledChanged = pyqtSignal(bool)
    gridEnabledChanged = pyqtSignal(bool)
    fontChanged = pyqtSignal(QFont)
    fontSizeChanged = pyqtSignal(int)

    def __init__(self, bargraph):
        super().__init__()

        self.m_graph = bargraph

        self.m_xRotation = 0.0
        self.m_yRotation = 0.0
        self.m_fontSize = 30
        self.m_segments = 4
        self.m_subSegments = 3
        self.m_minval = -20.0
        self.m_maxval = 20.0
        self.m_temperatureAxis = QValue3DAxis()
        self.m_yearAxis = QCategory3DAxis()
        self.m_monthAxis = QCategory3DAxis()
        self.m_primarySeries = QBar3DSeries()
        self.m_secondarySeries = QBar3DSeries()
        self.m_barMesh = QAbstract3DSeries.Mesh.MeshBevelBar
        self.m_smooth = False

        self.m_graph.setShadowQuality(
                QAbstract3DGraph.ShadowQuality.ShadowQualitySoftMedium)
        self.m_graph.activeTheme().setBackgroundEnabled(False)
        self.m_graph.activeTheme().setFont(
                QFont('Times New Roman', self.m_fontSize))
        self.m_graph.activeTheme().setLabelBackgroundEnabled(True)
        self.m_graph.setMultiSeriesUniform(True)

        self.m_temperatureAxis.setTitle("Average temperature")
        self.m_temperatureAxis.setSegmentCount(self.m_segments)
        self.m_temperatureAxis.setSubSegmentCount(self.m_subSegments)
        self.m_temperatureAxis.setRange(self.m_minval, self.m_maxval)
        self.m_temperatureAxis.setLabelFormat(u"%.1f \N{degree sign}C")

        self.m_yearAxis.setTitle("Year")
        self.m_monthAxis.setTitle("Month")

        self.m_graph.setValueAxis(self.m_temperatureAxis)
        self.m_graph.setRowAxis(self.m_yearAxis)
        self.m_graph.setColumnAxis(self.m_monthAxis)

        self.m_primarySeries.setItemLabelFormat(
                "Oulu - @colLabel @rowLabel: @valueLabel")
        self.m_primarySeries.setMesh(QAbstract3DSeries.Mesh.MeshBevelBar)
        self.m_primarySeries.setMeshSmooth(False)

        self.m_secondarySeries.setItemLabelFormat(
                "Helsinki - @colLabel @rowLabel: @valueLabel")
        self.m_secondarySeries.setMesh(QAbstract3DSeries.Mesh.MeshBevelBar)
        self.m_secondarySeries.setMeshSmooth(False)
        self.m_secondarySeries.setVisible(False)

        self.m_graph.addSeries(self.m_primarySeries)
        self.m_graph.addSeries(self.m_secondarySeries)

        self.m_preset = Q3DCamera.CameraPreset.CameraPresetFront
        self.changePresetCamera()

        self.resetTemperatureData()

    def resetTemperatureData(self):
        dataSet = []
        for row in self.tempOulu:
            dataSet.append([QBarDataItem(v) for v in row])

        self.m_primarySeries.dataProxy().resetArray(dataSet, self.years,
                self.months)

        dataSet = []
        for row in self.tempHelsinki:
            dataSet.append([QBarDataItem(v) for v in row])

        self.m_secondarySeries.dataProxy().resetArray(dataSet, self.years,
                self.months)

    def changeRange(self, range):
        if range >= len(self.years):
            self.m_yearAxis.setRange(0, len(self.years) - 1)
        else:
            self.m_yearAxis.setRange(range, range)

    def changeStyle(self, style):
        comboBox = self.sender()
        if isinstance(comboBox, QComboBox):
            self.m_barMesh = QAbstract3DSeries.Mesh(comboBox.itemData(style))
            self.m_primarySeries.setMesh(self.m_barMesh)
            self.m_secondarySeries.setMesh(self.m_barMesh)

    def changePresetCamera(self):
        self.m_graph.scene().activeCamera().setCameraPreset(self.m_preset)

        preset = int(self.m_preset.value) + 1
        if preset > Q3DCamera.CameraPreset.CameraPresetDirectlyBelow.value:
            self.m_preset = Q3DCamera.CameraPreset.CameraPresetFrontLow
        else:
            self.m_preset = Q3DCamera.CameraPreset(preset)

    def changeTheme(self, theme):
        currentTheme = self.m_graph.activeTheme()
        currentTheme.setType(Q3DTheme.Theme(theme))
        self.backgroundEnabledChanged.emit(currentTheme.isBackgroundEnabled())
        self.gridEnabledChanged.emit(currentTheme.isGridEnabled())
        self.fontChanged.emit(currentTheme.font())
        self.fontSizeChanged.emit(currentTheme.font().pointSize())

    def changeLabelBackground(self):
        self.m_graph.activeTheme().setLabelBackgroundEnabled(
                not self.m_graph.activeTheme().isLabelBackgroundEnabled())

    def changeSelectionMode(self, selectionMode):
        comboBox = selfsender()
        if isinstance(comboBox, QComboBox):
            flags = comboBox.itemData(selectionMode)
            self.m_graph.setSelectionMode(
                    QAbstract3DGraph.SelectionFlags(flags))

    def changeFont(self, font):
        self.m_graph.activeTheme().setFont(font)

    def changeFontSize(self, fontsize):
        self.m_fontSize = fontsize
        font = self.m_graph.activeTheme().font()
        font.setPointSize(self.m_fontSize)
        self.m_graph.activeTheme().setFont(font)

    def shadowQualityUpdatedByVisual(self, sq):
        self.shadowQualityChanged.emit(int(sq))

    def changeShadowQuality(self, quality):
        sq = QAbstract3DGraph.ShadowQuality(quality)
        self.m_graph.setShadowQuality(sq)
        self.shadowQualityChanged.emit(quality)

    def rotateX(self, rotation):
        self.m_xRotation = rotation
        self.m_graph.scene().activeCamera().setCameraPosition(
                self.m_xRotation, self.m_yRotation)

    def rotateY(self, rotation):
        self.m_yRotation = rotation
        self.m_graph.scene().activeCamera().setCameraPosition(
                self.m_xRotation, self.m_yRotation)

    def setBackgroundEnabled(self, enabled):
        self.m_graph.activeTheme().setBackgroundEnabled(enabled)

    def setGridEnabled(self, enabled):
        self.m_graph.activeTheme().setGridEnabled(enabled)

    def setSmoothBars(self, smooth):
        self.m_smooth = bool(smooth)
        self.m_primarySeries.setMeshSmooth(self.m_smooth)
        self.m_secondarySeries.setMeshSmooth(self.m_smooth)

    def setSeriesVisibility(self, enabled):
        self.m_secondarySeries.setVisible(enabled)


if __name__ == '__main__':

    import sys

    app = QApplication(sys.argv)
    widgetgraph = Q3DBars()
    container = QWidget.createWindowContainer(widgetgraph)

    screenSize = widgetgraph.screen().size()
    container.setMinimumSize(
            QSize(int(screenSize.width() / 2), int(screenSize.height() / 1.5)))
    container.setMaximumSize(screenSize)
    container.setSizePolicy(QSizePolicy.Policy.Expanding,
            QSizePolicy.Policy.Expanding)
    container.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

    widget = QWidget()
    hLayout = QHBoxLayout(widget)
    vLayout = QVBoxLayout()
    hLayout.addWidget(container, 1)
    hLayout.addLayout(vLayout)

    widget.setWindowTitle(
            "Average temperatures in Oulu and Helsinki, Finland (2006-2013)")

    themeList = QComboBox()
    themeList.addItem("Qt")
    themeList.addItem("Primary Colors")
    themeList.addItem("Digia")
    themeList.addItem("Stone Moss")
    themeList.addItem("Army Blue")
    themeList.addItem("Retro")
    themeList.addItem("Ebony")
    themeList.addItem("Isabelle")
    themeList.setCurrentIndex(0)

    labelButton = QPushButton("Change label style")

    smoothCheckBox = QCheckBox("Smooth bars")

    barStyleList = QComboBox()
    barStyleList.addItem("Bar", QAbstract3DSeries.Mesh.MeshBar)
    barStyleList.addItem("Pyramid", QAbstract3DSeries.Mesh.MeshPyramid)
    barStyleList.addItem("Cone", QAbstract3DSeries.Mesh.MeshCone)
    barStyleList.addItem("Cylinder", QAbstract3DSeries.Mesh.MeshCylinder)
    barStyleList.addItem("Bevel bar", QAbstract3DSeries.Mesh.MeshBevelBar)
    barStyleList.addItem("Sphere", QAbstract3DSeries.Mesh.MeshSphere)
    barStyleList.setCurrentIndex(4)

    cameraButton = QPushButton("Change camera preset")

    selectionModeList = QComboBox()
    selectionModeList.addItem("None",
            QAbstract3DGraph.SelectionFlag.SelectionNone)
    selectionModeList.addItem("Bar",
            QAbstract3DGraph.SelectionFlag.SelectionItem)
    selectionModeList.addItem("Row",
            QAbstract3DGraph.SelectionFlag.SelectionRow)
    selectionModeList.addItem("Bar and Row",
            QAbstract3DGraph.SelectionFlag.SelectionItemAndRow)
    selectionModeList.addItem("Column",
            QAbstract3DGraph.SelectionFlag.SelectionColumn)
    selectionModeList.addItem("Bar and Column",
            QAbstract3DGraph.SelectionFlag.SelectionItemAndColumn)
    selectionModeList.addItem("Row and Column",
            QAbstract3DGraph.SelectionFlag.SelectionRowAndColumn)
    selectionModeList.addItem("Bar, Row and Column",
            QAbstract3DGraph.SelectionFlag.SelectionItemRowAndColumn)
    selectionModeList.addItem("Slice into Row",
            QAbstract3DGraph.SelectionFlag.SelectionSlice | QAbstract3DGraph.SelectionFlag.SelectionRow)
    selectionModeList.addItem("Slice into Row and Item",
            QAbstract3DGraph.SelectionFlag.SelectionSlice | QAbstract3DGraph.SelectionFlag.SelectionItemAndRow)
    selectionModeList.addItem("Slice into Column",
            QAbstract3DGraph.SelectionFlag.SelectionSlice | QAbstract3DGraph.SelectionFlag.SelectionColumn)
    selectionModeList.addItem("Slice into Column and Item",
            QAbstract3DGraph.SelectionFlag.SelectionSlice | QAbstract3DGraph.SelectionFlag.SelectionItemAndColumn)
    selectionModeList.addItem("Multi: Bar, Row, Col",
            QAbstract3DGraph.SelectionFlag.SelectionItemRowAndColumn | QAbstract3DGraph.SelectionFlag.SelectionMultiSeries)
    selectionModeList.addItem("Multi, Slice: Row, Item",
            QAbstract3DGraph.SelectionFlag.SelectionSlice | QAbstract3DGraph.SelectionFlag.SelectionItemAndRow | QAbstract3DGraph.SelectionFlag.SelectionMultiSeries)
    selectionModeList.addItem("Multi, Slice: Col, Item",
            QAbstract3DGraph.SelectionFlag.SelectionSlice | QAbstract3DGraph.SelectionFlag.SelectionItemAndColumn | QAbstract3DGraph.SelectionFlag.SelectionMultiSeries)
    selectionModeList.setCurrentIndex(1)

    backgroundCheckBox = QCheckBox("Show background")

    gridCheckBox = QCheckBox("Show grid")
    gridCheckBox.setChecked(True)

    seriesCheckBox = QCheckBox("Show second series")

    rotationSliderX = QSlider(Qt.Orientation.Horizontal)
    rotationSliderX.setTickInterval(30)
    rotationSliderX.setTickPosition(QSlider.TickPosition.TicksBelow)
    rotationSliderX.setMinimum(-180)
    rotationSliderX.setValue(0)
    rotationSliderX.setMaximum(180)
    rotationSliderY = QSlider(Qt.Orientation.Horizontal)
    rotationSliderY.setTickInterval(15)
    rotationSliderY.setTickPosition(QSlider.TickPosition.TicksAbove)
    rotationSliderY.setMinimum(-90)
    rotationSliderY.setValue(0)
    rotationSliderY.setMaximum(90)

    fontSizeSlider = QSlider(Qt.Orientation.Horizontal)
    fontSizeSlider.setTickInterval(10)
    fontSizeSlider.setTickPosition(QSlider.TickPosition.TicksBelow)
    fontSizeSlider.setMinimum(1)
    fontSizeSlider.setValue(30)
    fontSizeSlider.setMaximum(100)

    fontList = QFontComboBox()
    fontList.setCurrentFont(QFont('Times New Roman'))

    shadowQuality = QComboBox()
    shadowQuality.addItem("None")
    shadowQuality.addItem("Low")
    shadowQuality.addItem("Medium")
    shadowQuality.addItem("High")
    shadowQuality.addItem("Low Soft")
    shadowQuality.addItem("Medium Soft")
    shadowQuality.addItem("High Soft")
    shadowQuality.setCurrentIndex(5)

    rangeList = QComboBox()
    rangeList.addItems(GraphModifier.years)
    rangeList.addItem("All")
    rangeList.setCurrentIndex(len(GraphModifier.years))

    vLayout.addWidget(QLabel("Rotate horizontally"))
    vLayout.addWidget(rotationSliderX, 0, Qt.AlignmentFlag.AlignTop)
    vLayout.addWidget(QLabel("Rotate vertically"))
    vLayout.addWidget(rotationSliderY, 0, Qt.AlignmentFlag.AlignTop)
    vLayout.addWidget(labelButton, 0, Qt.AlignmentFlag.AlignTop)
    vLayout.addWidget(cameraButton, 0, Qt.AlignmentFlag.AlignTop)
    vLayout.addWidget(backgroundCheckBox)
    vLayout.addWidget(gridCheckBox)
    vLayout.addWidget(smoothCheckBox)
    vLayout.addWidget(seriesCheckBox)
    vLayout.addWidget(QLabel("Show year"))
    vLayout.addWidget(rangeList)
    vLayout.addWidget(QLabel("Change bar style"))
    vLayout.addWidget(barStyleList)
    vLayout.addWidget(QLabel("Change selection mode"))
    vLayout.addWidget(selectionModeList)
    vLayout.addWidget(QLabel("Change theme"))
    vLayout.addWidget(themeList)
    vLayout.addWidget(QLabel("Adjust shadow quality"))
    vLayout.addWidget(shadowQuality)
    vLayout.addWidget(QLabel("Change font"))
    vLayout.addWidget(fontList)
    vLayout.addWidget(QLabel("Adjust font size"))
    vLayout.addWidget(fontSizeSlider, 1, Qt.AlignmentFlag.AlignTop)

    modifier = GraphModifier(widgetgraph)

    rotationSliderX.valueChanged.connect(modifier.rotateX)
    rotationSliderY.valueChanged.connect(modifier.rotateY)

    labelButton.clicked.connect(modifier.changeLabelBackground)
    cameraButton.clicked.connect(modifier.changePresetCamera)

    backgroundCheckBox.stateChanged.connect(modifier.setBackgroundEnabled)
    gridCheckBox.stateChanged.connect(modifier.setGridEnabled)
    smoothCheckBox.stateChanged.connect(modifier.setSmoothBars)
    seriesCheckBox.stateChanged.connect(modifier.setSeriesVisibility)

    modifier.backgroundEnabledChanged.connect(backgroundCheckBox.setChecked)
    modifier.gridEnabledChanged.connect(gridCheckBox.setChecked)

    rangeList.currentIndexChanged.connect(modifier.changeRange)

    barStyleList.currentIndexChanged.connect(modifier.changeStyle)

    selectionModeList.currentIndexChanged.connect(modifier.changeSelectionMode)

    themeList.currentIndexChanged.connect(modifier.changeTheme)

    shadowQuality.currentIndexChanged.connect(modifier.changeShadowQuality)

    modifier.shadowQualityChanged.connect(shadowQuality.setCurrentIndex)
    widgetgraph.shadowQualityChanged.connect(
            modifier.shadowQualityUpdatedByVisual)

    fontSizeSlider.valueChanged.connect(modifier.changeFontSize)
    fontList.currentFontChanged.connect(modifier.changeFont)

    modifier.fontSizeChanged.connect(fontSizeSlider.setValue)
    modifier.fontChanged.connect(fontList.setCurrentFont)

    widget.show()
    sys.exit(app.exec())    



#############################################################################
##
## Copyright (C) 2021 Riverbank Computing Limited.
## Copyright (C) 2014 Digia Plc
## All rights reserved.
## For any questions to Digia, please use contact form at http://qt.digia.com
##
## This file is part of the QtDataVisualization module.
##
## Licensees holding valid Qt Enterprise licenses may use this file in
## accordance with the Qt Enterprise License Agreement provided with the
## Software or, alternatively, in accordance with the terms contained in
## a written agreement between you and Digia.
##
## If you have questions regarding the use of this file, please use
## contact form at http://qt.digia.com
##
#############################################################################


from PyQt6.QtCore import (pyqtSignal, QFile, QFileInfo, QIODevice, QObject,
        QPoint, QSizeF, Qt, QTextStream)
from PyQt6.QtDataVisualization import (Q3DBars, Q3DCamera, Q3DTheme,
        QAbstract3DGraph, QAbstract3DSeries, QBar3DSeries, QBarDataItem,
        QBarDataProxy)
from PyQt6.QtGui import QFont, QGuiApplication


class VariantBarDataProxy(QBarDataProxy):

    def __init__(self, newSet=None, mapping=None):
        super().__init__()

        self.m_dataSet = None
        self.m_mapping = None

        if newSet is not None:
            self.setDataSet(newSet)

        if mapping is not None:
            self.setMapping(mapping)

    def setDataSet(self, newSet):
        if self.m_dataSet is not None:
            self.m_dataSet.itemsAdded.disconnect(self.handleItemsAdded)
            self.m_dataSet.dataCleared.disconnect(self.handleDataCleared)

        self.m_dataSet = newSet

        if self.m_dataSet is not None:
            self.m_dataSet.itemsAdded.connect(self.handleItemsAdded)
            self.m_dataSet.dataCleared.connect(self.handleDataCleared)

        self.resolveDataSet()

    def dataSet(self):
        return self.m_dataSet

    def setMapping(self, mapping):
        if self.m_mapping is not None:
            self.m_mapping.mappingChanged.disconnect(self.handleMappingChanged)

        self.m_mapping = mapping

        if self.m_mapping is not None:
            self.m_mapping.mappingChanged.connect(self.handleMappingChanged)

        self.resolveDataSet()

    def mapping(self):
        return self.m_mapping

    def handleItemsAdded(self, index, count):
        self.resolveDataSet()

    def handleDataCleared(self):
        self.resetArray(None)

    def handleMappingChanged(self):
        self.resolveDataSet()

    def resolveDataSet(self):
        # If we have no data or mapping, or the categories are not defined,
        # simply clear the array.
        if self.m_dataSet is None or self.m_mapping is None or len(self.m_mapping.rowCategories()) == 0 or len(self.m_mapping.columnCategories()) == 0:
            self.resetArray(None)
            return

        rowIndex = self.m_mapping.rowIndex()
        columnIndex = self.m_mapping.columnIndex()
        valueIndex = self.m_mapping.valueIndex()
        rowList = self.m_mapping.rowCategories()
        columnList = self.m_mapping.columnCategories()

        # Sort values into rows and columns.
        itemValueMap = {}
        for item in self.m_dataSet.itemList():
            columnValueMap = itemValueMap.setdefault(item[rowIndex], {})
            columnValueMap[item[columnIndex]] = item[valueIndex]

        newProxyArray = []
        for rowKey in rowList:
            newProxyRow = []
            for columnKey in columnList:
                newProxyRow.append(
                        QBarDataItem(itemValueMap[rowKey][columnKey]))

            newProxyArray.append(newProxyRow)

        self.resetArray(newProxyArray)


class VariantBarDataMapping(QObject):

    mappingChanged = pyqtSignal()

    def __init__(self, rowIndex=0, columnIndex=1, valueIndex=2,
            rowCategories=None, columnCategories=None):
        super().__init__()

        self.m_rowIndex = rowIndex
        self.m_columnIndex = columnIndex
        self.m_valueIndex = valueIndex

        self.m_rowCategories = [] if rowCategories is None else list(rowCategories)
        self.m_columnCategories = [] if columnCategories is None else list(columnCategories)

    def setRowIndex(self, index):
        self.m_rowIndex = index
        self.mappingChanged.emit()

    def rowIndex(self):
        return self.m_rowIndex

    def setColumnIndex(self, index):
        self.m_columnIndex = index
        self.mappingChanged.emit()

    def columnIndex(self):
        return self.m_columnIndex

    def setValueIndex(self, index):
        self.m_valueIndex = index
        self.mappingChanged.emit()

    def valueIndex(self):
        return self.m_valueIndex

    def setRowCategories(self, categories):
        self.m_rowCategories[:] = categories
        self.mappingChanged.emit()

    def rowCategories(self):
        return self.m_rowCategories

    def setColumnCategories(self, categories):
        self.m_columnCategories[:] = categories
        self.mappingChanged.emit()

    def columnCategories(self):
        return self.m_columnCategories

    def remap(sefl, rowIndex, columnIndex, valueIndex, rowCategories,
            columnCategories):
        self.m_rowIndex = rowIndex
        self.m_columnIndex = columnIndex
        self.m_valueIndex = valueIndex
        self.m_rowCategories[:] = rowCategories
        self.m_columnCategories[:] = columnCategories
        self.mappingChanged.emit()


class VariantDataSet(QObject):

    itemsAdded = pyqtSignal(int, int)
    dataCleared = pyqtSignal()

    def __init__(self):
        super().__init__()

        self.m_variantData = []

    def clear(self):
        del self.m_variantData[:]
        self.dataCleared.emit()

    def addItem(self, item):
        addIndex = len(self.m_variantData)
        self.m_variantData.append(item)

        self.itemsAdded.emit(addIndex, 1)

        return addIndex

    def addItems(self, itemList):
        addIndex = len(self.m_variantData)
        newCount = len(itemList)

        self.m_variantData += itemList

        self.itemsAdded.emit(addIndex, newCount)

        return addIndex

    def itemList(self):
        return self.m_variantData


class RainfallGraph(QObject):

    months = ("January", "February", "March", "April", "May", "June", "July",
        "August", "September", "October", "November", "December")

    def __init__(self, rainfall):
        super().__init__()

        self.m_graph = rainfall

        self.m_mapping = None
        self.m_dataSet = None

        # In the data file the months are in numeric format, so create a custom
        # list.
        self.m_numericMonths = [str(m) for m in range(1, 12 + 1)]
        self.m_columnCount = len(self.m_numericMonths)

        self.m_years = [str(y) for y in range(2000, 2012 + 1)]
        self.m_rowCount = len(self.m_years)

        self.m_proxy = VariantBarDataProxy()
        series = QBar3DSeries(self.m_proxy)
        series.setMesh(QAbstract3DSeries.Mesh.MeshCylinder)
        self.m_graph.addSeries(series)

        self.m_graph.setBarThickness(1.0)
        self.m_graph.setBarSpacing(QSizeF(0.2, 0.2))

        self.m_graph.rowAxis().setTitle("Year")
        self.m_graph.columnAxis().setTitle("Month")
        self.m_graph.valueAxis().setTitle("rainfall")
        self.m_graph.valueAxis().setLabelFormat("%d mm")
        self.m_graph.valueAxis().setSegmentCount(5)
        self.m_graph.rowAxis().setLabels(self.m_years)
        self.m_graph.columnAxis().setLabels(self.months)

        self.m_graph.setShadowQuality(
                QAbstract3DGraph.ShadowQuality.ShadowQualityMedium)

        self.m_graph.setSelectionMode(
                QAbstract3DGraph.SelectionFlag.SelectionItemAndColumn |
                QAbstract3DGraph.SelectionFlag.SelectionSlice)

        self.m_graph.activeTheme().setType(Q3DTheme.Theme.ThemeArmyBlue)
        self.m_graph.activeTheme().setFont(QFont('Century Gothic', 30))
        self.m_graph.activeTheme().setLabelBackgroundEnabled(False)

        self.m_graph.scene().activeCamera().setCameraPreset(
                Q3DCamera.CameraPreset.CameraPresetIsometricRightHigh)

        self.m_graph.setTitle("Monthly rainfall in Northern Finland")

    def start(self):
        self.addDataSet()

    def addDataSet(self):
        self.m_dataSet = VariantDataSet()
        itemList = []

        stream = QTextStream()
        dataFile = QFile(QFileInfo(__file__).absolutePath() + '/raindata.txt')
        if dataFile.open(QIODevice.OpenModeFlag.ReadOnly | QIODevice.OpenModeFlag.Text):
            stream.setDevice(dataFile)
            while not stream.atEnd():
                line = stream.readLine()
                if line.startswith('#'):
                    continue

                # Each line has three data items: year, month, and rainfall
                # values.
                strList = line.split(',')
                if len(strList) < 3:
                    continue

                # Store year and month as strings, and rainfall value as a
                # float into a tuple and add the tupe to the item list.
                newItem = (strList[0].strip(), strList[1].strip(),
                        float(strList[2]))
                itemList.append(newItem)

        self.m_dataSet.addItems(itemList)
        self.m_proxy.setDataSet(self.m_dataSet)

        self.m_mapping = VariantBarDataMapping(rowCategories=self.m_years,
                columnCategories=self.m_numericMonths)
        self.m_proxy.setMapping(self.m_mapping)


if __name__ == '__main__':

    import sys

    app = QGuiApplication(sys.argv)

    rainfall = Q3DBars()
    rainfall.setFlags(rainfall.flags() ^ Qt.WindowType.FramelessWindowHint)
    rainfall.resize(1000, 800)
    rainfall.setPosition(QPoint(10, 30))
    rainfall.show()

    rainfallgraph = RainfallGraph(rainfall)
    rainfallgraph.start()

    sys.exit(app.exec())



#############################################################################
##
## Copyright (C) 2021 Riverbank Computing Limited.
## Copyright (C) 2014 Digia Plc
## All rights reserved.
## For any questions to Digia, please use contact form at http://qt.digia.com
##
## This file is part of the QtDataVisualization module.
##
## Licensees holding valid Qt Enterprise licenses may use this file in
## accordance with the Qt Enterprise License Agreement provided with the
## Software or, alternatively, in accordance with the terms contained in
## a written agreement between you and Digia.
##
## If you have questions regarding the use of this file, please use
## contact form at http://qt.digia.com
##
#############################################################################


import math

from PyQt6.QtCore import pyqtSignal, QObject, QSize, Qt
from PyQt6.QtDataVisualization import (Q3DCamera, Q3DTheme, Q3DScatter,
        QAbstract3DGraph, QAbstract3DSeries, QScatter3DSeries,
        QScatterDataItem, QScatterDataProxy)
from PyQt6.QtGui import QFont, QVector3D
from PyQt6.QtWidgets import (QApplication, QCheckBox, QComboBox, QFontComboBox,
        QHBoxLayout, QLabel, QPushButton, QSizePolicy, QVBoxLayout, QWidget)


class ScatterDataModifier(QObject):

    numberOfItems = 3600
    curveDivider = 3.0
    lowerNumberOfItems = 900
    lowerCurveDivider = 0.75

    backgroundEnabledChanged = pyqtSignal(bool)
    gridEnabledChanged = pyqtSignal(bool)
    shadowQualityChanged = pyqtSignal(int)
    fontChanged = pyqtSignal(QFont)

    def __init__(self, scatter):
        super().__init__()

        self.m_graph = scatter
        self.m_fontSize = 40.0
        self.m_style = QAbstract3DSeries.Mesh.MeshSphere
        self.m_smooth = True
        self.m_itemCount = self.lowerNumberOfItems
        self.m_curveDivider = self.lowerCurveDivider
        self.m_preset = Q3DCamera.CameraPreset.CameraPresetFrontLow

        self.m_graph.activeTheme().setType(Q3DTheme.Theme.ThemeEbony)
        font = self.m_graph.activeTheme().font()
        font.setPointSizeF(self.m_fontSize)
        self.m_graph.activeTheme().setFont(font)
        self.m_graph.setShadowQuality(
                QAbstract3DGraph.ShadowQuality.ShadowQualitySoftLow)
        self.m_graph.scene().activeCamera().setCameraPreset(
                Q3DCamera.CameraPreset.CameraPresetFront)

        proxy = QScatterDataProxy()
        series = QScatter3DSeries(proxy)
        series.setItemLabelFormat(
                "@xTitle: @xLabel @yTitle: @yLabel @zTitle: @zLabel")
        series.setMeshSmooth(self.m_smooth)
        self.m_graph.addSeries(series)

        self.addData()

    def addData(self):
        self.m_graph.axisX().setTitle("X")
        self.m_graph.axisY().setTitle("Y")
        self.m_graph.axisZ().setTitle("Z")

        dataArray = []

        limit = math.sqrt(self.m_itemCount) / 2.0
        i = -limit
        while i < limit:
            j = -limit
            while j < limit:
                itm = QScatterDataItem(
                        QVector3D(i + 0.5,
                                math.cos(
                                        math.radians((i * j) / self.m_curveDivider)),
                                j + 0.5))
                dataArray.append(itm)
                j += 1.0

            i += 1.0

        self.m_graph.seriesList()[0].dataProxy().resetArray(dataArray)

    def changeStyle(self, style):
        comboBox = self.sender()
        if isinstance(comboBox, QComboBox):
            self.m_style = QAbstract3DSeries.Mesh(comboBox.itemData(style))
            self.m_graph.seriesList()[0].setMesh(self.m_style)

    def setSmoothDots(self, smooth):
        self.m_smooth = bool(smooth)
        self.m_graph.seriesList()[0].setMeshSmooth(self.m_smooth)

    def changeTheme(self, theme):
        currentTheme = self.m_graph.activeTheme()
        currentTheme.setType(Q3DTheme.Theme(theme))
        self.backgroundEnabledChanged.emit(currentTheme.isBackgroundEnabled())
        self.gridEnabledChanged.emit(currentTheme.isGridEnabled())
        self.fontChanged.emit(currentTheme.font())

    def changePresetCamera(self):
        self.m_graph.scene().activeCamera().setCameraPreset(self.m_preset)

        preset = self.m_preset.value + 1

        if preset > Q3DCamera.CameraPreset.CameraPresetDirectlyBelow.value:
            preset = Q3DCamera.CameraPreset.CameraPresetFrontLow.value

        self.m_preset = Q3DCamera.CameraPreset(preset)

    def changeLabelStyle(self):
        self.m_graph.activeTheme().setLabelBackgroundEnabled(
                not self.m_graph.activeTheme().isLabelBackgroundEnabled())

    def changeFont(self, font):
        newFont = QFont(font)
        newFont.setPointSizeF(self.m_fontSize)
        self.m_graph.activeTheme().setFont(newFont)

    def shadowQualityUpdatedByVisual(self, sq):
        self.shadowQualityChanged.emit(int(sq))

    def changeShadowQuality(self, quality):
        sq = QAbstract3DGraph.ShadowQuality(quality)
        self.m_graph.setShadowQuality(sq)

    def setBackgroundEnabled(self, enabled):
        self.m_graph.activeTheme().setBackgroundEnabled(enabled)

    def setGridEnabled(self, enabled):
        self.m_graph.activeTheme().setGridEnabled(enabled)

    def toggleItemCount(self):
        if self.m_itemCount == self.numberOfItems:
            self.m_itemCount = self.lowerNumberOfItems
            self.m_curveDivider = self.lowerCurveDivider
        else:
            self.m_itemCount = self.numberOfItems
            self.m_curveDivider = self.curveDivider

        self.m_graph.seriesList()[0].dataProxy().resetArray(None)
        self.addData()


if __name__ == '__main__':

    import sys

    app = QApplication(sys.argv)
    graph = Q3DScatter()
    container = QWidget.createWindowContainer(graph)

    screenSize = graph.screen().size()
    container.setMinimumSize(
            QSize(int(screenSize.width() / 2), int(screenSize.height() / 1.5)))
    container.setMaximumSize(screenSize)
    container.setSizePolicy(QSizePolicy.Policy.Expanding,
            QSizePolicy.Policy.Expanding)
    container.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

    widget = QWidget()
    hLayout = QHBoxLayout(widget)
    vLayout = QVBoxLayout()
    hLayout.addWidget(container, 1)
    hLayout.addLayout(vLayout)

    widget.setWindowTitle("A Cosine Wave")

    themeList = QComboBox()
    themeList.addItem("Qt")
    themeList.addItem("Primary Colors")
    themeList.addItem("Digia")
    themeList.addItem("Stone Moss")
    themeList.addItem("Army Blue")
    themeList.addItem("Retro")
    themeList.addItem("Ebony")
    themeList.addItem("Isabelle")
    themeList.setCurrentIndex(6)

    labelButton = QPushButton("Change label style")

    smoothCheckBox = QCheckBox("Smooth dots", checked=True)

    itemStyleList = QComboBox()
    itemStyleList.addItem("Sphere", QAbstract3DSeries.Mesh.MeshSphere)
    itemStyleList.addItem("Cube", QAbstract3DSeries.Mesh.MeshCube)
    itemStyleList.addItem("Minimal", QAbstract3DSeries.Mesh.MeshMinimal)
    itemStyleList.addItem("Point", QAbstract3DSeries.Mesh.MeshPoint)
    itemStyleList.setCurrentIndex(0)

    cameraButton = QPushButton("Change camera preset")

    itemCountButton = QPushButton("Toggle item count")

    backgroundCheckBox = QCheckBox("Show background", checked=True)

    gridCheckBox = QCheckBox("Show grid", checked=True)

    shadowQuality = QComboBox()
    shadowQuality.addItem("None")
    shadowQuality.addItem("Low")
    shadowQuality.addItem("Medium")
    shadowQuality.addItem("High")
    shadowQuality.addItem("Low Soft")
    shadowQuality.addItem("Medium Soft")
    shadowQuality.addItem("High Soft")
    shadowQuality.setCurrentIndex(4)

    fontList = QFontComboBox()
    fontList.setCurrentFont(QFont('Arial'))

    vLayout.addWidget(labelButton, 0, Qt.AlignmentFlag.AlignTop)
    vLayout.addWidget(cameraButton, 0, Qt.AlignmentFlag.AlignTop)
    vLayout.addWidget(itemCountButton, 0, Qt.AlignmentFlag.AlignTop)
    vLayout.addWidget(backgroundCheckBox)
    vLayout.addWidget(gridCheckBox)
    vLayout.addWidget(smoothCheckBox, 0, Qt.AlignmentFlag.AlignTop)
    vLayout.addWidget(QLabel("Change dot style"))
    vLayout.addWidget(itemStyleList)
    vLayout.addWidget(QLabel("Change theme"))
    vLayout.addWidget(themeList)
    vLayout.addWidget(QLabel("Adjust shadow quality"))
    vLayout.addWidget(shadowQuality)
    vLayout.addWidget(QLabel("Change font"))
    vLayout.addWidget(fontList, 1, Qt.AlignmentFlag.AlignTop)

    modifier = ScatterDataModifier(graph)

    cameraButton.clicked.connect(modifier.changePresetCamera)
    labelButton.clicked.connect(modifier.changeLabelStyle)
    itemCountButton.clicked.connect(modifier.toggleItemCount)

    backgroundCheckBox.stateChanged.connect(modifier.setBackgroundEnabled)
    gridCheckBox.stateChanged.connect(modifier.setGridEnabled)
    smoothCheckBox.stateChanged.connect(modifier.setSmoothDots)

    modifier.backgroundEnabledChanged.connect(backgroundCheckBox.setChecked)
    modifier.gridEnabledChanged.connect(gridCheckBox.setChecked)
    itemStyleList.currentIndexChanged.connect(modifier.changeStyle)

    themeList.currentIndexChanged.connect(modifier.changeTheme)

    shadowQuality.currentIndexChanged.connect(modifier.changeShadowQuality)

    modifier.shadowQualityChanged.connect(shadowQuality.setCurrentIndex)
    graph.shadowQualityChanged.connect(modifier.shadowQualityUpdatedByVisual)

    fontList.currentFontChanged.connect(modifier.changeFont)

    modifier.fontChanged.connect(fontList.setCurrentFont)

    widget.show()
    sys.exit(app.exec())



#############################################################################
##
## Copyright (C) 2021 Riverbank Computing Limited.
## Copyright (C) 2014 Digia Plc
## All rights reserved.
## For any questions to Digia, please use contact form at http://qt.digia.com
##
## This file is part of the QtDataVisualization module.
##
## Licensees holding valid Qt Enterprise licenses may use this file in
## accordance with the Qt Enterprise License Agreement provided with the
## Software or, alternatively, in accordance with the terms contained in
## a written agreement between you and Digia.
##
## If you have questions regarding the use of this file, please use
## contact form at http://qt.digia.com
##
#############################################################################


from PyQt6.QtCore import QFileInfo, QObject, QSize, Qt, QTimer
from PyQt6.QtDataVisualization import (Q3DCamera, Q3DScatter, Q3DTheme,
        QAbstract3DGraph, QCustom3DLabel, QCustom3DVolume)
from PyQt6.QtGui import (qBlue, qGreen, qRed, qRgba, QFont, QImage,
        QOpenGLContext, QPixmap, QQuaternion, QVector3D)
from PyQt6.QtWidgets import (QApplication, QCheckBox, QFrame, QGroupBox,
        QHBoxLayout, QLabel, QRadioButton, QSizePolicy, QSlider, QVBoxLayout,
        QWidget)


class VolumetricModifier(QObject):

    lowDetailSize = 128
    mediumDetailSize = 256
    highDetailSize = 512
    colorTableSize = 256
    layerDataSize = 512
    mineShaftDiameter = 1

    airColorIndex = 254
    mineShaftColorIndex = 255
    layerColorThickness = 60
    heightToColorDiv = 140
    magmaColorsMin = 0
    magmaColorsMax = layerColorThickness
    aboveWaterGroundColorsMin = magmaColorsMax + 1
    aboveWaterGroundColorsMax = aboveWaterGroundColorsMin + layerColorThickness
    underWaterGroundColorsMin = aboveWaterGroundColorsMax + 1
    underWaterGroundColorsMax = underWaterGroundColorsMin + layerColorThickness
    waterColorsMin = underWaterGroundColorsMax + 1
    waterColorsMax = waterColorsMin + layerColorThickness
    terrainTransparency = 12

    def __init__(self, scatter):
        super().__init__()

        self.m_graph = scatter
        self.m_volumeItem = None
        self.m_sliceIndexX = self.lowDetailSize // 2
        self.m_sliceIndexY = self.lowDetailSize // 4
        self.m_sliceIndexZ = self.lowDetailSize // 2
        self.m_slicingX = False
        self.m_slicingY = False
        self.m_slicingZ = False
        self.m_mediumDetailRB = None
        self.m_highDetailRB = None
        self.m_lowDetailData = None
        self.m_mediumDetailData = None
        self.m_highDetailData = None
        self.m_mediumDetailIndex = 0
        self.m_timer = QTimer()
        self.m_highDetailIndex = 0
        self.m_mediumDetailShaftIndex = 0
        self.m_highDetailShaftIndex = 0
        self.m_sliceSliderX = None
        self.m_sliceSliderY = None
        self.m_sliceSliderZ = None
        self.m_colorTable1 = []
        self.m_colorTable2 = []
        self.m_usingPrimaryTable = True
        self.m_sliceLabelX = None
        self.m_sliceLabelY = None
        self.m_sliceLabelZ = None
        self.m_magmaLayer = None
        self.m_waterLayer = None
        self.m_groundLayer = None
        self.m_mineShaftArray = ()

        self.m_graph.activeTheme().setType(Q3DTheme.Theme.ThemeQt)
        self.m_graph.setShadowQuality(
                QAbstract3DGraph.ShadowQuality.ShadowQualityNone)
        self.m_graph.scene().activeCamera().setCameraPreset(
                Q3DCamera.CameraPreset.CameraPresetFront)
        self.m_graph.setOrthoProjection(True)
        self.m_graph.activeTheme().setBackgroundEnabled(False)

        # Only allow zooming at the center and limit the zoom to 200% to avoid
        # clipping issues.
        self.m_graph.activeInputHandler().setZoomAtTargetEnabled(False)
        self.m_graph.scene().activeCamera().setMaxZoomLevel(200.0)

        self.toggleAreaAll(True)

        if QOpenGLContext.currentContext().isOpenGLES():
            # OpenGL ES2 doesn't support 3D textures, so show a warning label
            # instead.
            warningLabel = QCustom3DLabel(
                    "QCustom3DVolume is not supported with OpenGL ES2",
                    QFont(), QVector3D(0.0, 0.5, 0.0),
                    QVector3D(1.5, 1.5, 0.0), QQuaternion())
            warningLabel.setPositionAbsolute(True)
            warningLabel.setFacingCamera(True)
            self.m_graph.addCustomItem(warningLabel)

        else:
            self.m_lowDetailData = bytearray((self.lowDetailSize ** 3) // 2)
            self.m_mediumDetailData = bytearray(
                    (self.mediumDetailSize ** 3) // 2)
            self.m_highDetailData = bytearray((self.highDetailSize ** 3) // 2)

            heightmaps_dir = QFileInfo(__file__).absolutePath() + '/heightmaps'
            self.m_groundLayer = self.initHeightMap(
                    heightmaps_dir + '/layer_ground.png')
            self.m_waterLayer = self.initHeightMap(
                    heightmaps_dir + '/layer_water.png')
            self.m_magmaLayer = self.initHeightMap(
                    heightmaps_dir + '/layer_magma.png')

            self.initMineShaftArray()

            self.createVolume(self.lowDetailSize, 0, self.lowDetailSize,
                    self.m_lowDetailData)
            self.excavateMineShaft(self.lowDetailSize, 0,
                    len(self.m_mineShaftArray), self.m_lowDetailData)

            self.m_volumeItem = QCustom3DVolume()

            # Adjust the water level to zero with a minor tweak to the
            # y-coordinate position and scaling.
            axisX_min = self.m_graph.axisX().min()
            axisX_max = self.m_graph.axisX().max()
            axisY_min = self.m_graph.axisY().min()
            axisY_max = self.m_graph.axisY().max()
            axisZ_min = self.m_graph.axisZ().min()
            axisZ_max = self.m_graph.axisZ().max()

            self.m_volumeItem.setScaling(
                    QVector3D(axisX_max - axisX_min,
                            (axisY_max - axisY_min) * 0.91,
                            axisZ_max - axisZ_min))
            self.m_volumeItem.setPosition(
                    QVector3D((axisX_max + axisX_min) / 2.0,
                            -0.045 * (axisY_max - axisY_min) + (axisY_max + axisY_min) / 2.0,
                            (axisZ_max + axisZ_min) / 2.0))

            self.m_volumeItem.setScalingAbsolute(False)
            self.m_volumeItem.setTextureWidth(self.lowDetailSize)
            self.m_volumeItem.setTextureHeight(self.lowDetailSize // 2)
            self.m_volumeItem.setTextureDepth(self.lowDetailSize)
            self.m_volumeItem.setTextureFormat(QImage.Format.Format_Indexed8)
            self.m_volumeItem.setTextureData(self.m_lowDetailData)

            # Generate colour tables.  The alternate colour table just has grey
            # gradients for all terrain except water.
            self.m_colorTable1 = [None] * self.colorTableSize
            self.m_colorTable2 = [None] * self.colorTableSize

            for i in range(self.colorTableSize - 2):
                if i < self.magmaColorsMax:
                    color1 = qRgba(130 - (i * 2), 0, 0, 255)
                    color2 = qRgba(((i - self.aboveWaterGroundColorsMax) * 2),
                            ((i - self.aboveWaterGroundColorsMax) * 2),
                            ((i - self.aboveWaterGroundColorsMax) * 2), 255)
                elif i < self.aboveWaterGroundColorsMax:
                    color1 = qRgba((i - self.magmaColorsMax) * 4,
                            ((i - self.magmaColorsMax) * 2) + 120,
                            (i - self.magmaColorsMax) * 5,
                            self.terrainTransparency)
                    color2 = qRgba(((i - self.aboveWaterGroundColorsMax) * 2),
                            ((i - self.aboveWaterGroundColorsMax) * 2),
                            ((i - self.aboveWaterGroundColorsMax) * 2), 255)
                elif i < self.underWaterGroundColorsMax:
                    color1 = qRgba(
                            ((self.layerColorThickness - i - self.aboveWaterGroundColorsMax)) + 70,
                            ((self.layerColorThickness - i - self.aboveWaterGroundColorsMax) * 2) + 20,
                            ((self.layerColorThickness - i - self.aboveWaterGroundColorsMax)) + 50,
                            self.terrainTransparency)
                    color2 = qRgba(((i - self.aboveWaterGroundColorsMax) * 2),
                            ((i - self.aboveWaterGroundColorsMax) * 2),
                            ((i - self.aboveWaterGroundColorsMax) * 2),
                            self.terrainTransparency)
                elif i < self.waterColorsMax:
                    color1 = color2 = qRgba(0, 0,
                            ((i - self.underWaterGroundColorsMax) * 2) + 120,
                            self.terrainTransparency)
                else:
                    # Not used.
                    color1 = color2 = qRgba(0, 0, 0, 0)

                self.m_colorTable1[i] = color1
                self.m_colorTable2[i] = color2

            self.m_colorTable1[self.airColorIndex] = qRgba(0, 0, 0, 0)
            self.m_colorTable1[self.mineShaftColorIndex] = qRgba(50, 50, 50,
                    255)

            self.m_colorTable2[self.airColorIndex] = qRgba(0, 0, 0, 0)
            self.m_colorTable2[self.mineShaftColorIndex] = qRgba(255, 255, 0,
                    255)

            self.m_volumeItem.setColorTable(self.m_colorTable1)

            self.m_volumeItem.setSliceFrameGaps(QVector3D(0.01, 0.02, 0.01))
            self.m_volumeItem.setSliceFrameThicknesses(
                    QVector3D(0.0025, 0.005, 0.0025))
            self.m_volumeItem.setSliceFrameWidths(
                    QVector3D(0.0025, 0.005, 0.0025))
            self.m_volumeItem.setDrawSliceFrames(False);
            self.handleSlicingChanges();

            self.m_graph.addCustomItem(self.m_volumeItem);

            self.m_timer.start(0)

        self.m_graph.currentFpsChanged.connect(self.handleFpsChange)
        self.m_timer.timeout.connect(self.handleTimeout)

    def setFpsLabel(self, fpsLabel):
        self.m_fpsLabel = fpsLabel

    def setMediumDetailRB(self, button):
        self.m_mediumDetailRB = button

    def setHighDetailRB(self, button):
        self.m_highDetailRB = button

    def setSliceLabels(self, xLabel, yLabel, zLabel):
        self.m_sliceLabelX = xLabel
        self.m_sliceLabelY = yLabel
        self.m_sliceLabelZ = zLabel

        self.adjustSliceX(self.m_sliceSliderX.value())
        self.adjustSliceY(self.m_sliceSliderY.value())
        self.adjustSliceZ(self.m_sliceSliderZ.value())

    def setAlphaMultiplierLabel(self, label):
        self.m_alphaMultiplierLabel = label

    def sliceX(self, enabled):
        self.m_slicingX = enabled
        self.handleSlicingChanges()

    def sliceY(self, enabled):
        self.m_slicingY = enabled
        self.handleSlicingChanges()

    def sliceZ(self, enabled):
        self.m_slicingZ = enabled
        self.handleSlicingChanges()

    def adjustSliceX(self, value):
        if self.m_volumeItem is not None:
            self.m_sliceIndexX = value // (1024 // self.m_volumeItem.textureWidth())
            if self.m_sliceIndexX == self.m_volumeItem.textureWidth():
                self.m_sliceIndexX -= 1
            if self.m_volumeItem.sliceIndexX() != -1:
                self.m_volumeItem.setSliceIndexX(self.m_sliceIndexX)
            self.m_sliceLabelX.setPixmap(
                    QPixmap.fromImage(
                            self.m_volumeItem.renderSlice(Qt.Axis.XAxis,
                                    self.m_sliceIndexX)))

    def adjustSliceY(self, value):
        if self.m_volumeItem is not None:
            self.m_sliceIndexY = value // (1024 // self.m_volumeItem.textureHeight())
            if self.m_sliceIndexY == self.m_volumeItem.textureHeight():
                self.m_sliceIndexY -= 1
            if self.m_volumeItem.sliceIndexY() != -1:
                self.m_volumeItem.setSliceIndexY(self.m_sliceIndexY)
            self.m_sliceLabelY.setPixmap(
                    QPixmap.fromImage(
                            self.m_volumeItem.renderSlice(Qt.Axis.YAxis,
                                    self.m_sliceIndexY)))

    def adjustSliceZ(self, value):
        if self.m_volumeItem is not None:
            self.m_sliceIndexZ = value // (1024 // self.m_volumeItem.textureDepth())
            if self.m_sliceIndexZ == self.m_volumeItem.textureDepth():
                self.m_sliceIndexZ -= 1
            if self.m_volumeItem.sliceIndexZ() != -1:
                self.m_volumeItem.setSliceIndexZ(self.m_sliceIndexZ)
            self.m_sliceLabelZ.setPixmap(
                    QPixmap.fromImage(
                            self.m_volumeItem.renderSlice(Qt.Axis.ZAxis,
                                    self.m_sliceIndexZ)))

    def handleFpsChange(self, fps):
        self.m_fpsLabel.setText("FPS: %.1f" % fps)

    def handleTimeout(self):
        if not self.m_mediumDetailRB.isEnabled():
            if self.m_mediumDetailIndex != self.mediumDetailSize:
                self.m_mediumDetailIndex = self.createVolume(
                        self.mediumDetailSize, self.m_mediumDetailIndex, 4,
                        self.m_mediumDetailData)
            elif self.m_mediumDetailShaftIndex != len(self.m_mineShaftArray):
                self.m_mediumDetailShaftIndex = self.excavateMineShaft(
                        self.mediumDetailSize, self.m_mediumDetailShaftIndex,
                        1, self.m_mediumDetailData)
            else:
                self.m_mediumDetailRB.setEnabled(True)
                self.m_mediumDetailRB.setText(
                        "Medium (%dx%dx%d)" % (self.mediumDetailSize,
                                self.mediumDetailSize // 2,
                                self.mediumDetailSize))

        elif not self.m_highDetailRB.isEnabled():
            if self.m_highDetailIndex != self.highDetailSize:
                self.m_highDetailIndex = self.createVolume(
                        self.highDetailSize, self.m_highDetailIndex, 1,
                        self.m_highDetailData)
            elif self.m_highDetailShaftIndex != len(self.m_mineShaftArray):
                self.m_highDetailShaftIndex = self.excavateMineShaft(
                        self.highDetailSize, self.m_highDetailShaftIndex, 1,
                        self.m_highDetailData)
            else:
                self.m_highDetailRB.setEnabled(True)
                self.m_highDetailRB.setText(
                        "High (%dx%dx%d)" % (self.highDetailSize,
                                self.highDetailSize // 2, self.highDetailSize))
                self.m_timer.stop()

    def toggleLowDetail(self, enabled):
        if enabled and self.m_volumeItem is not None:
            self.m_volumeItem.setTextureData(self.m_lowDetailData)
            self.m_volumeItem.setTextureDimensions(self.lowDetailSize,
                    self.lowDetailSize // 2, self.lowDetailSize)
            self.adjustSliceX(self.m_sliceSliderX.value())
            self.adjustSliceY(self.m_sliceSliderY.value())
            self.adjustSliceZ(self.m_sliceSliderZ.value())

    def toggleMediumDetail(self, enabled):
        if enabled and self.m_volumeItem is not None:
            self.m_volumeItem.setTextureData(self.m_mediumDetailData)
            self.m_volumeItem.setTextureDimensions(self.mediumDetailSize,
                    self.mediumDetailSize // 2, self.mediumDetailSize)
            self.adjustSliceX(self.m_sliceSliderX.value())
            self.adjustSliceY(self.m_sliceSliderY.value())
            self.adjustSliceZ(self.m_sliceSliderZ.value())

    def toggleHighDetail(self, enabled):
        if enabled and self.m_volumeItem is not None:
            self.m_volumeItem.setTextureData(self.m_highDetailData)
            self.m_volumeItem.setTextureDimensions(self.highDetailSize,
                    self.highDetailSize // 2, self.highDetailSize)
            self.adjustSliceX(self.m_sliceSliderX.value())
            self.adjustSliceY(self.m_sliceSliderY.value())
            self.adjustSliceZ(self.m_sliceSliderZ.value())

    def setFpsMeasurement(self, enabled):
        self.m_graph.setMeasureFps(enabled)

        if enabled:
            self.m_fpsLabel.setText("Measuring...")
        else:
            self.m_fpsLabel.setText('')

    def setSliceSliders(self, sliderX, sliderY, sliderZ):
        self.m_sliceSliderX = sliderX
        self.m_sliceSliderY = sliderY
        self.m_sliceSliderZ = sliderZ

        # Set sliders to interesting values.
        self.m_sliceSliderX.setValue(715)
        self.m_sliceSliderY.setValue(612)
        self.m_sliceSliderZ.setValue(715)

    def changeColorTable(self, enabled):
        if self.m_volumeItem is not None:
            if enabled:
                self.m_volumeItem.setColorTable(self.m_colorTable2)
            else:
                self.m_volumeItem.setColorTable(self.m_colorTable1)

            self.m_usingPrimaryTable = not enabled

            # Re-render image labels.
            self.adjustSliceX(self.m_sliceSliderX.value())
            self.adjustSliceY(self.m_sliceSliderY.value())
            self.adjustSliceZ(self.m_sliceSliderZ.value())

    def setPreserveOpacity(self, enabled):
        if self.m_volumeItem is not None:
            self.m_volumeItem.setPreserveOpacity(enabled)

            # Re-render image labels.
            self.adjustSliceX(self.m_sliceSliderX.value())
            self.adjustSliceY(self.m_sliceSliderY.value())
            self.adjustSliceZ(self.m_sliceSliderZ.value())

    def setTransparentGround(self, enabled):
        if self.m_volumeItem is not None:
            newAlpha = self.terrainTransparency if enabled else 255

            for i in range(self.aboveWaterGroundColorsMin, self.underWaterGroundColorsMax):
                oldColor1 = self.m_colorTable1[i]
                oldColor2 = self.m_colorTable2[i]
                self.m_colorTable1[i] = qRgba(qRed(oldColor1), qGreen(oldColor1), qBlue(oldColor1), newAlpha)
                self.m_colorTable2[i] = qRgba(qRed(oldColor2), qGreen(oldColor2), qBlue(oldColor2), newAlpha)

            if self.m_usingPrimaryTable:
                self.m_volumeItem.setColorTable(self.m_colorTable1)
            else:
                self.m_volumeItem.setColorTable(self.m_colorTable2)

            self.adjustSliceX(self.m_sliceSliderX.value())
            self.adjustSliceY(self.m_sliceSliderY.value())
            self.adjustSliceZ(self.m_sliceSliderZ.value())

    def setUseHighDefShader(self, enabled):
        if self.m_volumeItem is not None:
            self.m_volumeItem.setUseHighDefShader(enabled)

    def adjustAlphaMultiplier(self, value):
        if self.m_volumeItem is not None:
            if value > 100:
                mult = (value - 99) / 2.0
            else:
                mult = value / float(500 - value * 4)

            self.m_volumeItem.setAlphaMultiplier(mult)
            self.m_alphaMultiplierLabel.setText("Alpha multiplier: %.3f" % mult)

            # Re-render image labels.
            self.adjustSliceX(self.m_sliceSliderX.value())
            self.adjustSliceY(self.m_sliceSliderY.value())
            self.adjustSliceZ(self.m_sliceSliderZ.value())

    def toggleAreaAll(self, enabled):
        if enabled:
            self.m_graph.axisX().setRange(0.0, 1000.0)
            self.m_graph.axisY().setRange(-600.0, 600.0)
            self.m_graph.axisZ().setRange(0.0, 1000.0)
            self.m_graph.axisX().setSegmentCount(5)
            self.m_graph.axisY().setSegmentCount(6)
            self.m_graph.axisZ().setSegmentCount(5)

    def toggleAreaMine(self, enabled):
        if enabled:
            self.m_graph.axisX().setRange(350.0, 850.0)
            self.m_graph.axisY().setRange(-500.0, 100.0)
            self.m_graph.axisZ().setRange(350.0, 900.0)
            self.m_graph.axisX().setSegmentCount(10)
            self.m_graph.axisY().setSegmentCount(6)
            self.m_graph.axisZ().setSegmentCount(11)

    def toggleAreaMountain(self, enabled):
        if enabled:
            self.m_graph.axisX().setRange(300.0, 600.0)
            self.m_graph.axisY().setRange(-100.0, 400.0)
            self.m_graph.axisZ().setRange(300.0, 600.0)
            self.m_graph.axisX().setSegmentCount(9)
            self.m_graph.axisY().setSegmentCount(5)
            self.m_graph.axisZ().setSegmentCount(9)

    def setDrawSliceFrames(self, enabled):
        if self.m_volumeItem is not None:
            self.m_volumeItem.setDrawSliceFrames(enabled)

    def initHeightMap(self, fileName):
        heightImage = QImage(fileName)
        bits = heightImage.constBits().asarray(heightImage.sizeInBytes())
        colorTable = heightImage.colorTable()

        layerData = bytearray(self.layerDataSize ** 2)
        index = 0

        for i in range(self.layerDataSize):
            for j in range(self.layerDataSize):
                layerData[index] = qRed(colorTable[bits[index]])
                index += 1

        return layerData

    def createVolume(self, textureSize, startIndex, count, textureData):
        # Generate volume from layer data.
        index = (startIndex * textureSize * textureSize) // 2
        endIndex = min(startIndex + count, textureSize)
        magmaHeights = bytearray(textureSize)
        waterHeights = bytearray(textureSize)
        groundHeights = bytearray(textureSize)
        multiplier = float(self.layerDataSize) / float(textureSize)

        for i in range(startIndex, endIndex):
            # Generate layer height arrays.
            for l in range(textureSize):
                layerIndex = int(i * multiplier) * self.layerDataSize + int(l * multiplier)
                magmaHeights[l] = self.m_magmaLayer[layerIndex]
                waterHeights[l] = self.m_waterLayer[layerIndex]
                groundHeights[l] = self.m_groundLayer[layerIndex]

            for j in range(textureSize // 2):
                for k in range(textureSize):
                    height = int((self.layerDataSize - (j * 2 * multiplier)) / 2)

                    if height < magmaHeights[k]:
                        # Magma layer.
                        colorIndex = int((float(height) / self.heightToColorDiv)
                                        * float(self.layerColorThickness)) + self.magmaColorsMin
                    elif height < groundHeights[k] and height < waterHeights[k]:
                        # Ground layer below water.
                        colorIndex = int((float(waterHeights[k] - height) / self.heightToColorDiv)
                                        * float(self.layerColorThickness)) + self.underWaterGroundColorsMin
                    elif height < waterHeights[k]:
                        # Water layer where water goes over ground.
                        colorIndex = int((float(height - magmaHeights[k]) / self.heightToColorDiv)
                                        * float(self.layerColorThickness)) + self.waterColorsMin
                    elif height <= groundHeights[k]:
                        # Ground above water.
                        colorIndex = int((float(height - waterHeights[k]) / self.heightToColorDiv)
                                        * float(self.layerColorThickness)) + self.aboveWaterGroundColorsMin
                    else:
                        # Rest is air.
                        colorIndex = self.airColorIndex

                    textureData[index] = colorIndex
                    index += 1

        return endIndex

    def excavateMineShaft(self, textureSize, startIndex, count, textureData):
        endIndex = min(startIndex + count, len(self.m_mineShaftArray))
        shaftSize = (self.mineShaftDiameter * textureSize) // self.lowDetailSize

        for shaftStart, shaftEnd in self.m_mineShaftArray[startIndex:endIndex]:
            shaftLen = int((shaftEnd - shaftStart).length() * self.lowDetailSize)
            dataX = int(shaftStart.x() * textureSize - (shaftSize // 2))
            dataY = int((shaftStart.y() * textureSize - (shaftSize // 2)) / 2)
            dataZ = int(shaftStart.z() * textureSize - (shaftSize // 2))
            dataIndex = dataX + (dataY * textureSize) + dataZ * (textureSize * textureSize // 2)

            if shaftStart.x() != shaftEnd.x():
                for j in range(shaftLen + 1):
                    self.excavateMineBlock(textureSize, dataIndex, shaftSize,
                            textureData)
                    dataIndex += shaftSize
            elif shaftStart.y() != shaftEnd.y():
                # Vertical shafts are half as long.
                shaftLen //= 2
                for j in range(shaftLen + 1):
                    self.excavateMineBlock(textureSize, dataIndex, shaftSize,
                            textureData)
                    dataIndex += textureSize * shaftSize
            else:
                for j in range(shaftLen + 1):
                    self.excavateMineBlock(textureSize, dataIndex, shaftSize,
                            textureData)
                    dataIndex += (textureSize * textureSize // 2) * shaftSize

        return endIndex

    @classmethod
    def excavateMineBlock(cls, textureSize, dataIndex, size, textureData):
        for k in range(size):
            for l in range(size):
                curIndex = dataIndex + ((k * textureSize * textureSize) // 2) + (l * textureSize)
                for m in range(size):
                    if textureData[curIndex] != cls.airColorIndex:
                        textureData[curIndex] = cls.mineShaftColorIndex
                    curIndex += 1

    def handleSlicingChanges(self):
        if self.m_volumeItem is not None:
            if self.m_slicingX or self.m_slicingY or self.m_slicingZ:
                # Only show slices of selected dimensions.
                self.m_volumeItem.setDrawSlices(True)
                self.m_volumeItem.setSliceIndexX(self.m_sliceIndexX if self.m_slicingX else -1)
                self.m_volumeItem.setSliceIndexY(self.m_sliceIndexY if self.m_slicingY else -1)
                self.m_volumeItem.setSliceIndexZ(self.m_sliceIndexZ if self.m_slicingZ else -1)
            else:
                # Show slice frames for all dimensions when not actually
                # slicing.
                self.m_volumeItem.setDrawSlices(False)
                self.m_volumeItem.setSliceIndexX(self.m_sliceIndexX)
                self.m_volumeItem.setSliceIndexY(self.m_sliceIndexY)
                self.m_volumeItem.setSliceIndexZ(self.m_sliceIndexZ)

    def initMineShaftArray(self):
        self.m_mineShaftArray = (
            (QVector3D(0.7, 0.1, 0.7), QVector3D(0.7, 0.8, 0.7)),
            (QVector3D(0.7, 0.7, 0.5), QVector3D(0.7, 0.7, 0.7)),
            (QVector3D(0.4, 0.7, 0.7), QVector3D(0.7, 0.7, 0.7)),
            (QVector3D(0.4, 0.7, 0.7), QVector3D(0.4, 0.7, 0.8)),
            (QVector3D(0.45, 0.7, 0.7), QVector3D(0.45, 0.7, 0.8)),
            (QVector3D(0.5, 0.7, 0.7), QVector3D(0.5, 0.7, 0.8)),
            (QVector3D(0.55, 0.7, 0.7), QVector3D(0.55, 0.7, 0.8)),
            (QVector3D(0.6, 0.7, 0.7), QVector3D(0.6, 0.7, 0.8)),
            (QVector3D(0.65, 0.7, 0.7), QVector3D(0.65, 0.7, 0.8)),
            (QVector3D(0.5, 0.6, 0.7), QVector3D(0.7, 0.6, 0.7)),
            (QVector3D(0.5, 0.6, 0.7), QVector3D(0.5, 0.6, 0.8)),
            (QVector3D(0.55, 0.6, 0.7), QVector3D(0.55, 0.6, 0.8)),
            (QVector3D(0.6, 0.6, 0.7), QVector3D(0.6, 0.6, 0.8)),
            (QVector3D(0.65, 0.6, 0.7), QVector3D(0.65, 0.6, 0.8)),
            (QVector3D(0.7, 0.6, 0.4), QVector3D(0.7, 0.6, 0.7)),
            (QVector3D(0.6, 0.6, 0.45), QVector3D(0.8, 0.6, 0.45)),
            (QVector3D(0.6, 0.6, 0.5), QVector3D(0.8, 0.6, 0.5)),
            (QVector3D(0.6, 0.6, 0.55), QVector3D(0.8, 0.6, 0.55)),
            (QVector3D(0.6, 0.6, 0.6), QVector3D(0.8, 0.6, 0.6)),
            (QVector3D(0.6, 0.6, 0.65), QVector3D(0.8, 0.6, 0.65)),
            (QVector3D(0.6, 0.6, 0.7), QVector3D(0.8, 0.6, 0.7)),
            (QVector3D(0.7, 0.7, 0.4), QVector3D(0.7, 0.7, 0.7)),
            (QVector3D(0.6, 0.7, 0.45), QVector3D(0.8, 0.7, 0.45)),
            (QVector3D(0.6, 0.7, 0.5), QVector3D(0.8, 0.7, 0.5)),
            (QVector3D(0.6, 0.7, 0.55), QVector3D(0.8, 0.7, 0.55)),
            (QVector3D(0.6, 0.7, 0.6), QVector3D(0.8, 0.7, 0.6)),
            (QVector3D(0.6, 0.7, 0.65), QVector3D(0.8, 0.7, 0.65)),
            (QVector3D(0.6, 0.7, 0.7), QVector3D(0.8, 0.7, 0.7)),
            (QVector3D(0.7, 0.8, 0.5), QVector3D(0.7, 0.8, 0.7)),
            (QVector3D(0.6, 0.8, 0.55), QVector3D(0.8, 0.8, 0.55)),
            (QVector3D(0.6, 0.8, 0.6), QVector3D(0.8, 0.8, 0.6)),
            (QVector3D(0.6, 0.8, 0.65), QVector3D(0.8, 0.8, 0.65)),
            (QVector3D(0.6, 0.8, 0.7), QVector3D(0.8, 0.8, 0.7)),
            (QVector3D(0.7, 0.1, 0.4), QVector3D(0.7, 0.7, 0.4)))


if __name__ == '__main__':

    import sys

    app = QApplication(sys.argv)

    graph = Q3DScatter()
    container = QWidget.createWindowContainer(graph)

    screenSize = graph.screen().size()
    container.setMinimumSize(
            QSize(screenSize.width() // 3, screenSize.height() // 3))
    container.setMaximumSize(screenSize)
    container.setSizePolicy(QSizePolicy.Policy.Expanding,
            QSizePolicy.Policy.Expanding)
    container.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

    widget = QWidget()
    hLayout = QHBoxLayout(widget)
    vLayout = QVBoxLayout()
    vLayout2 = QVBoxLayout()
    hLayout.addWidget(container, 1)
    hLayout.addLayout(vLayout)
    hLayout.addLayout(vLayout2)

    widget.setWindowTitle("Volumetric object example - 3D terrain")

    sliceXCheckBox = QCheckBox()
    sliceXCheckBox.setText("Slice volume on X axis")
    sliceXCheckBox.setChecked(False)
    sliceYCheckBox = QCheckBox()
    sliceYCheckBox.setText("Slice volume on Y axis")
    sliceYCheckBox.setChecked(False)
    sliceZCheckBox = QCheckBox()
    sliceZCheckBox.setText("Slice volume on Z axis")
    sliceZCheckBox.setChecked(False)

    sliceXSlider = QSlider(Qt.Orientation.Horizontal)
    sliceXSlider.setMinimum(0)
    sliceXSlider.setMaximum(1024)
    sliceXSlider.setValue(512)
    sliceXSlider.setEnabled(True)
    sliceYSlider = QSlider(Qt.Orientation.Horizontal)
    sliceYSlider.setMinimum(0)
    sliceYSlider.setMaximum(1024)
    sliceYSlider.setValue(512)
    sliceYSlider.setEnabled(True)
    sliceZSlider = QSlider(Qt.Orientation.Horizontal)
    sliceZSlider.setMinimum(0)
    sliceZSlider.setMaximum(1024)
    sliceZSlider.setValue(512)
    sliceZSlider.setEnabled(True)

    fpsCheckBox = QCheckBox()
    fpsCheckBox.setText("Show FPS")
    fpsCheckBox.setChecked(False)
    fpsLabel = QLabel()

    textureDetailGroupBox = QGroupBox("Texture detail")

    lowDetailRB = QRadioButton()
    lowDetailRB.setText("Low (128x64x128)")
    lowDetailRB.setChecked(True)

    mediumDetailRB = QRadioButton()
    mediumDetailRB.setText("Generating...")
    mediumDetailRB.setChecked(False)
    mediumDetailRB.setEnabled(False)

    highDetailRB = QRadioButton()
    highDetailRB.setText("Generating...")
    highDetailRB.setChecked(False)
    highDetailRB.setEnabled(False)

    textureDetailVBox = QVBoxLayout()
    textureDetailVBox.addWidget(lowDetailRB)
    textureDetailVBox.addWidget(mediumDetailRB)
    textureDetailVBox.addWidget(highDetailRB)
    textureDetailGroupBox.setLayout(textureDetailVBox)

    areaGroupBox = QGroupBox("Show area")

    areaAllRB = QRadioButton()
    areaAllRB.setText("Whole region")
    areaAllRB.setChecked(True)

    areaMineRB = QRadioButton()
    areaMineRB.setText("The mine")
    areaMineRB.setChecked(False)

    areaMountainRB = QRadioButton()
    areaMountainRB.setText("The mountain")
    areaMountainRB.setChecked(False)

    areaVBox = QVBoxLayout()
    areaVBox.addWidget(areaAllRB)
    areaVBox.addWidget(areaMineRB)
    areaVBox.addWidget(areaMountainRB)
    areaGroupBox.setLayout(areaVBox)

    colorTableCheckBox = QCheckBox()
    colorTableCheckBox.setText("Alternate color table")
    colorTableCheckBox.setChecked(False)

    sliceImageXLabel = QLabel()
    sliceImageYLabel = QLabel()
    sliceImageZLabel = QLabel()
    sliceImageXLabel.setMinimumSize(QSize(200, 100))
    sliceImageYLabel.setMinimumSize(QSize(200, 200))
    sliceImageZLabel.setMinimumSize(QSize(200, 100))
    sliceImageXLabel.setMaximumSize(QSize(200, 100))
    sliceImageYLabel.setMaximumSize(QSize(200, 200))
    sliceImageZLabel.setMaximumSize(QSize(200, 100))
    sliceImageXLabel.setFrameShape(QFrame.Shape.Box)
    sliceImageYLabel.setFrameShape(QFrame.Shape.Box)
    sliceImageZLabel.setFrameShape(QFrame.Shape.Box)
    sliceImageXLabel.setScaledContents(True)
    sliceImageYLabel.setScaledContents(True)
    sliceImageZLabel.setScaledContents(True)

    alphaMultiplierSlider = QSlider(Qt.Orientation.Horizontal)
    alphaMultiplierSlider.setMinimum(0)
    alphaMultiplierSlider.setMaximum(139)
    alphaMultiplierSlider.setValue(100)
    alphaMultiplierSlider.setEnabled(True)
    alphaMultiplierLabel = QLabel("Alpha multiplier: 1.0")

    preserveOpacityCheckBox = QCheckBox()
    preserveOpacityCheckBox.setText("Preserve opacity")
    preserveOpacityCheckBox.setChecked(True)

    transparentGroundCheckBox = QCheckBox()
    transparentGroundCheckBox.setText("Transparent ground")
    transparentGroundCheckBox.setChecked(False)

    useHighDefShaderCheckBox = QCheckBox()
    useHighDefShaderCheckBox.setText("Use HD shader")
    useHighDefShaderCheckBox.setChecked(True)

    performanceNoteLabel = QLabel(
            "Note: A high end graphics card is\nrecommended with the HD shader\nwhen the volume contains a lot of\ntransparent areas.")
    performanceNoteLabel.setFrameShape(QFrame.Shape.Box)

    drawSliceFramesCheckBox = QCheckBox()
    drawSliceFramesCheckBox.setText("Draw slice frames")
    drawSliceFramesCheckBox.setChecked(False)

    vLayout.addWidget(sliceXCheckBox)
    vLayout.addWidget(sliceXSlider)
    vLayout.addWidget(sliceImageXLabel)
    vLayout.addWidget(sliceYCheckBox)
    vLayout.addWidget(sliceYSlider)
    vLayout.addWidget(sliceImageYLabel)
    vLayout.addWidget(sliceZCheckBox)
    vLayout.addWidget(sliceZSlider)
    vLayout.addWidget(sliceImageZLabel)
    vLayout.addWidget(drawSliceFramesCheckBox, 1, Qt.AlignmentFlag.AlignTop)
    vLayout2.addWidget(fpsCheckBox)
    vLayout2.addWidget(fpsLabel)
    vLayout2.addWidget(textureDetailGroupBox)
    vLayout2.addWidget(areaGroupBox)
    vLayout2.addWidget(colorTableCheckBox)
    vLayout2.addWidget(alphaMultiplierLabel)
    vLayout2.addWidget(alphaMultiplierSlider)
    vLayout2.addWidget(preserveOpacityCheckBox)
    vLayout2.addWidget(transparentGroundCheckBox)
    vLayout2.addWidget(useHighDefShaderCheckBox)
    vLayout2.addWidget(performanceNoteLabel, 1, Qt.AlignmentFlag.AlignTop)

    modifier = VolumetricModifier(graph)
    modifier.setFpsLabel(fpsLabel)
    modifier.setMediumDetailRB(mediumDetailRB)
    modifier.setHighDetailRB(highDetailRB)
    modifier.setSliceSliders(sliceXSlider, sliceYSlider, sliceZSlider)
    modifier.setSliceLabels(sliceImageXLabel, sliceImageYLabel,
            sliceImageZLabel)
    modifier.setAlphaMultiplierLabel(alphaMultiplierLabel)
    modifier.setTransparentGround(transparentGroundCheckBox.isChecked())

    fpsCheckBox.stateChanged.connect(modifier.setFpsMeasurement)
    sliceXCheckBox.stateChanged.connect(modifier.sliceX)
    sliceYCheckBox.stateChanged.connect(modifier.sliceY)
    sliceZCheckBox.stateChanged.connect(modifier.sliceZ)
    sliceXSlider.valueChanged.connect(modifier.adjustSliceX)
    sliceYSlider.valueChanged.connect(modifier.adjustSliceY)
    sliceZSlider.valueChanged.connect(modifier.adjustSliceZ)
    lowDetailRB.toggled.connect(modifier.toggleLowDetail)
    mediumDetailRB.toggled.connect(modifier.toggleMediumDetail)
    highDetailRB.toggled.connect(modifier.toggleHighDetail)
    colorTableCheckBox.stateChanged.connect(modifier.changeColorTable)
    preserveOpacityCheckBox.stateChanged.connect(modifier.setPreserveOpacity)
    transparentGroundCheckBox.stateChanged.connect(
            modifier.setTransparentGround)
    useHighDefShaderCheckBox.stateChanged.connect(modifier.setUseHighDefShader)
    alphaMultiplierSlider.valueChanged.connect(modifier.adjustAlphaMultiplier)
    areaAllRB.toggled.connect(modifier.toggleAreaAll)
    areaMineRB.toggled.connect(modifier.toggleAreaMine)
    areaMountainRB.toggled.connect(modifier.toggleAreaMountain)
    drawSliceFramesCheckBox.stateChanged.connect(modifier.setDrawSliceFrames)

    widget.show()

    sys.exit(app.exec())



#############################################################################
##
## Copyright (C) 2021 Riverbank Computing Limited.
## Copyright (C) 2014 Digia Plc
## All rights reserved.
## For any questions to Digia, please use contact form at http://qt.digia.com
##
## This file is part of the QtDataVisualization module.
##
## Licensees holding valid Qt Enterprise licenses may use this file in
## accordance with the Qt Enterprise License Agreement provided with the
## Software or, alternatively, in accordance with the terms contained in
## a written agreement between you and Digia.
##
## If you have questions regarding the use of this file, please use
## contact form at http://qt.digia.com
##
#############################################################################


import math

from PyQt6.QtCore import QFileInfo, QObject, QSize, Qt, QTimer
from PyQt6.QtDataVisualization import (Q3DCamera, Q3DScatter, Q3DTheme,
        QAbstract3DGraph, QAbstract3DSeries, QScatter3DSeries,
        QScatterDataItem)
from PyQt6.QtGui import QLinearGradient, QQuaternion, QVector3D
from PyQt6.QtWidgets import (QApplication, QHBoxLayout, QLabel, QPushButton,
        QSlider, QSizePolicy, QVBoxLayout, QWidget)


class ScatterDataModifier(QObject):

    verticalRange = 8.0
    horizontalRange = verticalRange
    ellipse_a = horizontalRange / 3.0
    ellipse_b = verticalRange
    doublePi = math.pi * 2.0
    radiansToDegrees = 360.0 / doublePi
    animationFrames = 30.0

    def __init__(self, scatter):
        super().__init__()

        mesh_dir = QFileInfo(__file__).absolutePath() + '/mesh'

        self.m_graph = scatter
        self.m_rotationTimer = QTimer()
        self.m_fieldLines = 12
        self.m_arrowsPerLine = 16
        self.m_magneticField = QScatter3DSeries()
        self.m_sun = QScatter3DSeries()
        self.m_angleOffset = 0.0
        self.m_angleStep = self.doublePi / self.m_arrowsPerLine / self.animationFrames

        self.m_graph.setShadowQuality(
                QAbstract3DGraph.ShadowQuality.ShadowQualityNone)
        self.m_graph.scene().activeCamera().setCameraPreset(
                Q3DCamera.CameraPreset.CameraPresetFront)

        # Magnetic field lines use custom narrow arrow.
        self.m_magneticField.setItemSize(0.2)
        self.m_magneticField.setMesh(
                QAbstract3DSeries.Mesh.MeshUserDefined)
        self.m_magneticField.setUserDefinedMesh(mesh_dir + '/narrowarrow.obj')
        fieldGradient = QLinearGradient(0, 0, 16, 1024)
        fieldGradient.setColorAt(0.0, Qt.GlobalColor.black)
        fieldGradient.setColorAt(1.0, Qt.GlobalColor.white)
        self.m_magneticField.setBaseGradient(fieldGradient)
        self.m_magneticField.setColorStyle(
                Q3DTheme.ColorStyle.ColorStyleRangeGradient)

        # For 'sun' we use a custom large sphere.
        self.m_sun.setItemSize(0.2)
        self.m_sun.setName("Sun")
        self.m_sun.setItemLabelFormat("@seriesName")
        self.m_sun.setMesh(QAbstract3DSeries.Mesh.MeshUserDefined)
        self.m_sun.setUserDefinedMesh(mesh_dir + '/largesphere.obj')
        self.m_sun.setBaseColor(0xffbb00)
        self.m_sun.dataProxy().addItem(QScatterDataItem(QVector3D()))

        self.m_graph.addSeries(self.m_magneticField)
        self.m_graph.addSeries(self.m_sun)

        # Configure the axes according to the data.
        self.m_graph.axisX().setRange(-self.horizontalRange,
                self.horizontalRange)
        self.m_graph.axisY().setRange(-self.verticalRange, self.verticalRange)
        self.m_graph.axisZ().setRange(-self.horizontalRange,
                self.horizontalRange)
        self.m_graph.axisX().setSegmentCount(int(self.horizontalRange))
        self.m_graph.axisZ().setSegmentCount(int(self.horizontalRange))

        self.m_rotationTimer.timeout.connect(self.triggerRotation)

        self.toggleRotation()
        self.generateData()

    def generateData(self):
        magneticFieldArray = []

        for i in range(self.m_fieldLines):
            horizontalAngle = (self.doublePi * i) / self.m_fieldLines
            xCenter = self.ellipse_a * math.cos(horizontalAngle)
            zCenter = self.ellipse_a * math.sin(horizontalAngle)

            # Rotate - arrow is always tangential to the origin.
            yRotation = QQuaternion.fromAxisAndAngle(0.0, 1.0, 0.0,
                    horizontalAngle * self.radiansToDegrees)

            for j in range(self.m_arrowsPerLine):
                # Calculate the point on the ellipse centered on the origin and
                # parallel to the x-axis.
                verticalAngle = ((self.doublePi * j) / self.m_arrowsPerLine) + self.m_angleOffset
                xUnrotated = self.ellipse_a * math.cos(verticalAngle)
                y = self.ellipse_b * math.sin(verticalAngle)

                # Rotate the ellipse around the y-axis.
                xRotated = xUnrotated * math.cos(horizontalAngle)
                zRotated = xUnrotated * math.sin(horizontalAngle)

                # Add the offset.
                x = xCenter + xRotated
                z = zCenter + zRotated

                zRotation = QQuaternion.fromAxisAndAngle(0.0, 0.0, 1.0,
                        verticalAngle * self.radiansToDegrees)
                totalRotation = yRotation * zRotation

                itm = QScatterDataItem(QVector3D(x, y, z), totalRotation)
                magneticFieldArray.append(itm)

        if self.m_graph.selectedSeries() is self.m_magneticField:
            self.m_graph.clearSelection()

        self.m_magneticField.dataProxy().resetArray(magneticFieldArray)

    def setFieldLines(self, lines):
        self.m_fieldLines = lines
        self.generateData()

    def setArrowsPerLine(self, arrows):
        self.m_arrowsPerLine = arrows
        self.m_angleOffset = 0.0
        self.m_angleStep = self.doublePi / self.m_arrowsPerLine / self.animationFrames
        self.generateData()

    def triggerRotation(self):
        self.m_angleOffset += self.m_angleStep
        self.generateData()

    def toggleSun(self):
        self.m_sun.setVisible(not self.m_graph.seriesList()[1].isVisible())

    def toggleRotation(self):
        if self.m_rotationTimer.isActive():
            self.m_rotationTimer.stop()
        else:
            self.m_rotationTimer.start(15)


if __name__ == '__main__':

    import sys

    app = QApplication(sys.argv)
    graph = Q3DScatter()
    container = QWidget.createWindowContainer(graph)

    screenSize = graph.screen().size()
    container.setMinimumSize(
            QSize(int(screenSize.width() / 2), int(screenSize.height() / 1.5)))
    container.setMaximumSize(screenSize)
    container.setSizePolicy(QSizePolicy.Policy.Expanding,
            QSizePolicy.Policy.Expanding)
    container.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

    widget = QWidget()
    hLayout = QHBoxLayout(widget)
    vLayout = QVBoxLayout()
    hLayout.addWidget(container, 1)
    hLayout.addLayout(vLayout)

    widget.setWindowTitle("Item rotations example - Magnetic field of the sun")

    toggleRotationButton = QPushButton("Toggle animation")
    toggleSunButton = QPushButton("Toggle Sun")

    fieldLinesSlider = QSlider(Qt.Orientation.Horizontal)
    fieldLinesSlider.setTickInterval(1)
    fieldLinesSlider.setMinimum(1)
    fieldLinesSlider.setValue(12)
    fieldLinesSlider.setMaximum(128)

    arrowsSlider = QSlider(Qt.Orientation.Horizontal)
    arrowsSlider.setTickInterval(1)
    arrowsSlider.setMinimum(8)
    arrowsSlider.setValue(16)
    arrowsSlider.setMaximum(32)

    vLayout.addWidget(toggleRotationButton)
    vLayout.addWidget(toggleSunButton)
    vLayout.addWidget(QLabel("Field Lines (1 - 128):"))
    vLayout.addWidget(fieldLinesSlider)
    vLayout.addWidget(QLabel("Arrows per line (8 - 32):"))
    vLayout.addWidget(arrowsSlider, 1, Qt.AlignmentFlag.AlignTop)

    modifier = ScatterDataModifier(graph)

    toggleRotationButton.clicked.connect(modifier.toggleRotation)
    toggleSunButton.clicked.connect(modifier.toggleSun)
    fieldLinesSlider.valueChanged.connect(modifier.setFieldLines)
    arrowsSlider.valueChanged.connect(modifier.setArrowsPerLine)

    widget.show()
    sys.exit(app.exec())



#############################################################################
##
## Copyright (C) 2021 Riverbank Computing Limited.
## Copyright (C) 2014 Digia Plc
## All rights reserved.
## For any questions to Digia, please use contact form at http://qt.digia.com
##
## This file is part of the QtDataVisualization module.
##
## Licensees holding valid Qt Enterprise licenses may use this file in
## accordance with the Qt Enterprise License Agreement provided with the
## Software or, alternatively, in accordance with the terms contained in
## a written agreement between you and Digia.
##
## If you have questions regarding the use of this file, please use
## contact form at http://qt.digia.com
##
#############################################################################


import math

from PyQt6.QtCore import QFileInfo, QObject, QSize, Qt
from PyQt6.QtDataVisualization import (Q3DSurface, Q3DTheme, QAbstract3DGraph,
        QHeightMapSurfaceDataProxy, QSurface3DSeries, QSurfaceDataItem,
        QSurfaceDataProxy, QValue3DAxis)
from PyQt6.QtGui import (QBrush, QIcon, QImage, QLinearGradient, QPainter,
        QPixmap, QVector3D)
from PyQt6.QtWidgets import (QApplication, QComboBox, QGroupBox, QHBoxLayout,
        QLabel, QPushButton, QRadioButton, QSizePolicy, QSlider, QVBoxLayout,
        QWidget)


class SurfaceGraph(QObject):

    sampleCountX = 50
    sampleCountZ = 50
    heightMapGridStepX = 6
    heightMapGridStepZ = 6
    sampleMin = -8.0
    sampleMax = 8.0

    def __init__(self, surface):
        super().__init__()

        self.m_graph = surface
        self.m_graph.setAxisX(QValue3DAxis())
        self.m_graph.setAxisY(QValue3DAxis())
        self.m_graph.setAxisZ(QValue3DAxis())

        self.m_sqrtSinProxy = QSurfaceDataProxy()
        self.m_sqrtSinSeries = QSurface3DSeries(self.m_sqrtSinProxy)
        self.fillSqrtSinProxy()

        heightMapImage = QImage(
                QFileInfo(__file__).absolutePath() + '/mountain.png')
        self.m_heightMapProxy = QHeightMapSurfaceDataProxy(heightMapImage)
        self.m_heightMapSeries = QSurface3DSeries(self.m_heightMapProxy)
        self.m_heightMapSeries.setItemLabelFormat("(@xLabel, @zLabel): @yLabel")
        self.m_heightMapProxy.setValueRanges(34.0, 40.0, 18.0, 24.0)
        self.m_heightMapWidth = heightMapImage.width()
        self.m_heightMapHeight = heightMapImage.height()

        self.m_axisMinSliderX = None
        self.m_axisMaxSliderX = None
        self.m_axisMinSliderZ = None
        self.m_axisMaxSliderZ = None
        self.m_rangeMinX = 0.0
        self.m_rangeMinZ = 0.0
        self.m_stepX = 0.0
        self.m_stepZ = 0.0

    def fillSqrtSinProxy(self):
        stepX = (self.sampleMax - self.sampleMin) / (self.sampleCountX - 1)
        stepZ = (self.sampleMax - self.sampleMin) / (self.sampleCountZ - 1)

        dataArray = []
        for i in range(self.sampleCountZ):

            # Keep values within range bounds, since just adding step can cause
            # minor drift due to the rounding errors.
            z = min(self.sampleMax, (i * stepZ + self.sampleMin))
            newRow = []
            for j in range(self.sampleCountX):
                x = min(self.sampleMax, (j * stepX + self.sampleMin))
                R = math.sqrt(z * z + x * x) + 0.01
                y = (math.sin(R) / R + 0.24) * 1.61

                newRow.append(QSurfaceDataItem(QVector3D(x, y, z)))

            dataArray.append(newRow)

        self.m_sqrtSinProxy.resetArray(dataArray)

    def toggleModeNone(self):
        self.m_graph.setSelectionMode(
                QAbstract3DGraph.SelectionFlag.SelectionNone)

    def toggleModeItem(self):
        self.m_graph.setSelectionMode(
                QAbstract3DGraph.SelectionFlag.SelectionItem)

    def toggleModeSliceRow(self):
        self.m_graph.setSelectionMode(
                QAbstract3DGraph.SelectionFlag.SelectionItemAndRow |
                QAbstract3DGraph.SelectionFlag.SelectionSlice)

    def toggleModeSliceColumn(self):
        self.m_graph.setSelectionMode(
                QAbstract3DGraph.SelectionFlag.SelectionItemAndColumn |
                QAbstract3DGraph.SelectionFlag.SelectionSlice)

    def setAxisMinSliderX(self, slider):
        self.m_axisMinSliderX = slider

    def setAxisMaxSliderX(self, slider):
        self.m_axisMaxSliderX = slider

    def setAxisMinSliderZ(self, slider):
        self.m_axisMinSliderZ = slider

    def setAxisMaxSliderZ(self, slider):
        self.m_axisMaxSliderZ = slider

    def enableSqrtSinModel(self, enable):
        if not enable:
            return

        self.m_sqrtSinSeries.setDrawMode(
                QSurface3DSeries.DrawFlag.DrawSurfaceAndWireframe)
        self.m_sqrtSinSeries.setFlatShadingEnabled(True)

        self.m_graph.axisX().setLabelFormat("%.2f")
        self.m_graph.axisZ().setLabelFormat("%.2f")
        self.m_graph.axisX().setRange(self.sampleMin, self.sampleMax)
        self.m_graph.axisY().setRange(0.0, 2.0)
        self.m_graph.axisZ().setRange(self.sampleMin, self.sampleMax)
        self.m_graph.axisX().setLabelAutoRotation(30)
        self.m_graph.axisY().setLabelAutoRotation(90)
        self.m_graph.axisZ().setLabelAutoRotation(30)

        self.m_graph.removeSeries(self.m_heightMapSeries)
        self.m_graph.addSeries(self.m_sqrtSinSeries)

        self.m_rangeMinX = self.sampleMin
        self.m_rangeMinZ = self.sampleMin
        self.m_stepX = (self.sampleMax - self.sampleMin) / (self.sampleCountX - 1)
        self.m_stepZ = (self.sampleMax - self.sampleMin) / (self.sampleCountZ - 1)
        self.m_axisMinSliderX.setMaximum(self.sampleCountX - 2)
        self.m_axisMinSliderX.setValue(0)
        self.m_axisMaxSliderX.setMaximum(self.sampleCountX - 1)
        self.m_axisMaxSliderX.setValue(self.sampleCountX - 1)
        self.m_axisMinSliderZ.setMaximum(self.sampleCountZ - 2)
        self.m_axisMinSliderZ.setValue(0)
        self.m_axisMaxSliderZ.setMaximum(self.sampleCountZ - 1)
        self.m_axisMaxSliderZ.setValue(self.sampleCountZ - 1)

    def enableHeightMapModel(self, enable):
        if not enable:
            return

        self.m_heightMapSeries.setDrawMode(
                QSurface3DSeries.DrawFlag.DrawSurface)
        self.m_heightMapSeries.setFlatShadingEnabled(False)

        self.m_graph.axisX().setLabelFormat("%.1f N")
        self.m_graph.axisZ().setLabelFormat("%.1f E")
        self.m_graph.axisX().setRange(34.0, 40.0)
        self.m_graph.axisY().setAutoAdjustRange(True)
        self.m_graph.axisZ().setRange(18.0, 24.0)

        self.m_graph.axisX().setTitle("Latitude")
        self.m_graph.axisY().setTitle("Height")
        self.m_graph.axisZ().setTitle("Longitude")

        self.m_graph.removeSeries(self.m_sqrtSinSeries)
        self.m_graph.addSeries(self.m_heightMapSeries)

        mapGridCountX = self.m_heightMapWidth // self.heightMapGridStepX
        mapGridCountZ = self.m_heightMapHeight // self.heightMapGridStepZ
        self.m_rangeMinX = 34.0
        self.m_rangeMinZ = 18.0
        self.m_stepX = 6.0 / (mapGridCountX - 1)
        self.m_stepZ = 6.0 / (mapGridCountZ - 1)
        self.m_axisMinSliderX.setMaximum(mapGridCountX - 2)
        self.m_axisMinSliderX.setValue(0)
        self.m_axisMaxSliderX.setMaximum(mapGridCountX - 1)
        self.m_axisMaxSliderX.setValue(mapGridCountX - 1)
        self.m_axisMinSliderZ.setMaximum(mapGridCountZ - 2)
        self.m_axisMinSliderZ.setValue(0)
        self.m_axisMaxSliderZ.setMaximum(mapGridCountZ - 1)
        self.m_axisMaxSliderZ.setValue(mapGridCountZ - 1)

    def adjustXMin(self, min):
        minX = self.m_stepX * min + self.m_rangeMinX

        max = self.m_axisMaxSliderX.value()
        if min >= max:
            max = min + 1
            self.m_axisMaxSliderX.setValue(max)

        maxX = self.m_stepX * max + self.m_rangeMinX

        self.setAxisXRange(minX, maxX)

    def adjustXMax(self, max):
        maxX = self.m_stepX * max + self.m_rangeMinX

        min = self.m_axisMinSliderX.value()
        if max <= min:
            min = max - 1
            self.m_axisMinSliderX.setValue(min)

        minX = self.m_stepX * min + self.m_rangeMinX

        self.setAxisXRange(minX, maxX)

    def adjustZMin(self, min):
        minZ = self.m_stepZ * min + self.m_rangeMinZ

        max = self.m_axisMaxSliderZ.value()
        if min >= max:
            max = min + 1
            self.m_axisMaxSliderZ.setValue(max)

        maxZ = self.m_stepZ * max + self.m_rangeMinZ

        self.setAxisZRange(minZ, maxZ)

    def adjustZMax(self, max):
        maxZ = self.m_stepZ * max + self.m_rangeMinZ

        min = self.m_axisMinSliderZ.value()
        if max <= min:
            min = max - 1
            self.m_axisMinSliderZ.setValue(min)

        minZ = self.m_stepZ * min + self.m_rangeMinZ

        self.setAxisZRange(minZ, maxZ)

    def setAxisXRange(self, min, max):
        self.m_graph.axisX().setRange(min, max)

    def setAxisZRange(self, min, max):
        self.m_graph.axisZ().setRange(min, max)

    def changeTheme(self, theme):
        self.m_graph.activeTheme().setType(Q3DTheme.Theme(theme))

    def setBlackToYellowGradient(self):
        gr = QLinearGradient()
        gr.setColorAt(0.0, Qt.GlobalColor.black)
        gr.setColorAt(0.33, Qt.GlobalColor.blue)
        gr.setColorAt(0.67, Qt.GlobalColor.red)
        gr.setColorAt(1.0, Qt.GlobalColor.yellow)

        self.m_graph.seriesList()[0].setBaseGradient(gr)
        self.m_graph.seriesList()[0].setColorStyle(
                Q3DTheme.ColorStyle.ColorStyleRangeGradient)

    def setGreenToRedGradient(self):
        gr = QLinearGradient()
        gr.setColorAt(0.0, Qt.GlobalColor.darkGreen)
        gr.setColorAt(0.5, Qt.GlobalColor.yellow)
        gr.setColorAt(0.8, Qt.GlobalColor.red)
        gr.setColorAt(1.0, Qt.GlobalColor.darkRed)

        self.m_graph.seriesList()[0].setBaseGradient(gr)
        self.m_graph.seriesList()[0].setColorStyle(
                Q3DTheme.ColorStyle.ColorStyleRangeGradient)


if __name__ == '__main__':

    import sys

    app = QApplication(sys.argv)
    graph = Q3DSurface()
    container = QWidget.createWindowContainer(graph)

    screenSize = graph.screen().size()
    container.setMinimumSize(
            QSize(int(screenSize.width() / 2), int(screenSize.height() / 1.6)))
    container.setMaximumSize(screenSize)
    container.setSizePolicy(QSizePolicy.Policy.Expanding,
            QSizePolicy.Policy.Expanding)
    container.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

    widget = QWidget()
    hLayout = QHBoxLayout(widget)
    vLayout = QVBoxLayout()
    hLayout.addWidget(container, 1)
    hLayout.addLayout(vLayout)

    widget.setWindowTitle("Surface example")

    modelGroupBox = QGroupBox("Model")

    sqrtSinModelRB = QRadioButton("Sqrt && Sin")
    heightMapModelRB = QRadioButton("Height Map")

    modelVBox = QVBoxLayout()
    modelVBox.addWidget(sqrtSinModelRB)
    modelVBox.addWidget(heightMapModelRB)
    modelGroupBox.setLayout(modelVBox)

    selectionGroupBox = QGroupBox("Selection Mode")

    modeNoneRB = QRadioButton("No selection")
    modeItemRB = QRadioButton("Item")
    modeSliceRowRB = QRadioButton("Row Slice")
    modeSliceColumnRB = QRadioButton("Column Slice")

    selectionVBox = QVBoxLayout()
    selectionVBox.addWidget(modeNoneRB)
    selectionVBox.addWidget(modeItemRB)
    selectionVBox.addWidget(modeSliceRowRB)
    selectionVBox.addWidget(modeSliceColumnRB)
    selectionGroupBox.setLayout(selectionVBox)

    axisMinSliderX = QSlider(Qt.Orientation.Horizontal, minimum=0,
            tickInterval=1)
    axisMaxSliderX = QSlider(Qt.Orientation.Horizontal, minimum=1,
            tickInterval=1)
    axisMinSliderZ = QSlider(Qt.Orientation.Horizontal, minimum=0,
            tickInterval=1)
    axisMaxSliderZ = QSlider(Qt.Orientation.Horizontal, minimum=1,
            tickInterval=1)

    themeList = QComboBox()
    themeList.addItem("Qt")
    themeList.addItem("Primary Colors")
    themeList.addItem("Digia")
    themeList.addItem("Stone Moss")
    themeList.addItem("Army Blue")
    themeList.addItem("Retro")
    themeList.addItem("Ebony")
    themeList.addItem("Isabelle")

    colorGroupBox = QGroupBox("Custom gradient")

    grBtoY = QLinearGradient(0, 0, 1, 100)
    grBtoY.setColorAt(1.0, Qt.GlobalColor.black)
    grBtoY.setColorAt(0.67, Qt.GlobalColor.blue)
    grBtoY.setColorAt(0.33, Qt.GlobalColor.red)
    grBtoY.setColorAt(0.0, Qt.GlobalColor.yellow)
    pm = QPixmap(24, 100)
    pmp = QPainter(pm)
    pmp.setBrush(QBrush(grBtoY))
    pmp.setPen(Qt.PenStyle.NoPen)
    pmp.drawRect(0, 0, 24, 100)
    gradientBtoYPB = QPushButton()
    gradientBtoYPB.setIcon(QIcon(pm))
    gradientBtoYPB.setIconSize(QSize(24, 100))

    grGtoR = QLinearGradient(0, 0, 1, 100)
    grGtoR.setColorAt(1.0, Qt.GlobalColor.darkGreen)
    grGtoR.setColorAt(0.5, Qt.GlobalColor.yellow)
    grGtoR.setColorAt(0.2, Qt.GlobalColor.red)
    grGtoR.setColorAt(0.0, Qt.GlobalColor.darkRed)
    pmp.setBrush(QBrush(grGtoR))
    pmp.drawRect(0, 0, 24, 100)
    gradientGtoRPB = QPushButton()
    gradientGtoRPB.setIcon(QIcon(pm))
    gradientGtoRPB.setIconSize(QSize(24, 100))

    colorHBox = QHBoxLayout()
    colorHBox.addWidget(gradientBtoYPB)
    colorHBox.addWidget(gradientGtoRPB)
    colorGroupBox.setLayout(colorHBox)

    vLayout.addWidget(modelGroupBox)
    vLayout.addWidget(selectionGroupBox)
    vLayout.addWidget(QLabel("Column range"))
    vLayout.addWidget(axisMinSliderX)
    vLayout.addWidget(axisMaxSliderX)
    vLayout.addWidget(QLabel("Row range"))
    vLayout.addWidget(axisMinSliderZ)
    vLayout.addWidget(axisMaxSliderZ)
    vLayout.addWidget(QLabel("Theme"))
    vLayout.addWidget(themeList)
    vLayout.addWidget(colorGroupBox)

    widget.show()

    modifier = SurfaceGraph(graph)

    heightMapModelRB.toggled.connect(modifier.enableHeightMapModel)
    sqrtSinModelRB.toggled.connect(modifier.enableSqrtSinModel)
    modeNoneRB.toggled.connect(modifier.toggleModeNone)
    modeItemRB.toggled.connect(modifier.toggleModeItem)
    modeSliceRowRB.toggled.connect(modifier.toggleModeSliceRow)
    modeSliceColumnRB.toggled.connect(modifier.toggleModeSliceColumn)
    axisMinSliderX.valueChanged.connect(modifier.adjustXMin)
    axisMaxSliderX.valueChanged.connect(modifier.adjustXMax)
    axisMinSliderZ.valueChanged.connect(modifier.adjustZMin)
    axisMaxSliderZ.valueChanged.connect(modifier.adjustZMax)
    themeList.currentIndexChanged.connect(modifier.changeTheme)
    gradientBtoYPB.pressed.connect(modifier.setBlackToYellowGradient)
    gradientGtoRPB.pressed.connect(modifier.setGreenToRedGradient)

    modifier.setAxisMinSliderX(axisMinSliderX)
    modifier.setAxisMaxSliderX(axisMaxSliderX)
    modifier.setAxisMinSliderZ(axisMinSliderZ)
    modifier.setAxisMaxSliderZ(axisMaxSliderZ)

    sqrtSinModelRB.setChecked(True)
    modeItemRB.setChecked(True)
    themeList.setCurrentIndex(2)

    sys.exit(app.exec())



#############################################################################
##
## Copyright (C) 2021 Riverbank Computing Limited.
## Copyright (C) 2014 Digia Plc
## All rights reserved.
## For any questions to Digia, please use contact form at http://qt.digia.com
##
## This file is part of the QtDataVisualization module.
##
## Licensees holding valid Qt Enterprise licenses may use this file in
## accordance with the Qt Enterprise License Agreement provided with the
## Software or, alternatively, in accordance with the terms contained in
## a written agreement between you and Digia.
##
## If you have questions regarding the use of this file, please use
## contact form at http://qt.digia.com
##
#############################################################################


from PyQt6.QtCore import (pyqtSignal, QAbstractAnimation, QFile, QFileInfo,
        QIODevice, QObject, QPropertyAnimation, QSequentialAnimationGroup,
        QSize, Qt, QTextStream, QTimer)
from PyQt6.QtDataVisualization import (Q3DCamera, Q3DScatter, Q3DTheme,
        QAbstract3DGraph, QAbstract3DInputHandler, QAbstract3DSeries,
        QScatter3DSeries, QScatterDataItem, QValue3DAxis)
from PyQt6.QtGui import QVector3D
from PyQt6.QtWidgets import (QApplication, QComboBox, QHBoxLayout, QLabel,
        QPushButton, QSizePolicy, QVBoxLayout, QWidget)


class CustomInputHandler(QAbstract3DInputHandler):

    def mouseMoveEvent(self, event, mousePos):
        self.setInputPosition(mousePos)

    def wheelEvent(self, event):
        # Adjust the zoom level based on what zoom range we're in.
        zoomLevel = self.scene().activeCamera().zoomLevel()

        if zoomLevel > 100:
            zoomLevel += event.angleDelta().y() / 12
        elif zoomLevel > 50:
            zoomLevel += event.angleDelta().y() / 60
        else:
            zoomLevel += event.angleDelta().y() / 120

        if zoomLevel > 500:
            zoomLevel = 500
        elif zoomLevel < 10:
            zoomLevel = 10

        self.scene().activeCamera().setZoomLevel(zoomLevel)


class ScatterDataModifier(QObject):

    shadowQualityChanged = pyqtSignal()

    def __init__(self, scatter):
        super().__init__()

        self.m_graph = scatter
        self.m_inputHandler = CustomInputHandler()

        self.m_graph.activeTheme().setType(Q3DTheme.Theme.ThemeDigia)
        self.m_graph.setShadowQuality(
                QAbstract3DGraph.ShadowQuality.ShadowQualityMedium)
        self.m_graph.scene().activeCamera().setCameraPreset(
                Q3DCamera.CameraPreset.CameraPresetFront)

        self.m_graph.setAxisX(QValue3DAxis())
        self.m_graph.setAxisY(QValue3DAxis())
        self.m_graph.setAxisZ(QValue3DAxis())

        self.m_graph.axisX().setRange(-10.0, 10.0)
        self.m_graph.axisY().setRange(-5.0, 5.0)
        self.m_graph.axisZ().setRange(-5.0, 5.0)

        series = QScatter3DSeries()
        series.setItemLabelFormat("@xLabel, @yLabel, @zLabel")
        series.setMesh(QAbstract3DSeries.Mesh.MeshCube)
        series.setItemSize(0.15)
        self.m_graph.addSeries(series)

        self.m_animationCameraX = QPropertyAnimation(
                self.m_graph.scene().activeCamera(), b'xRotation')
        self.m_animationCameraX.setDuration(20000)
        self.m_animationCameraX.setStartValue(0.0)
        self.m_animationCameraX.setEndValue(360.0)
        self.m_animationCameraX.setLoopCount(-1)

        upAnimation = QPropertyAnimation(self.m_graph.scene().activeCamera(),
                b'yRotation')
        upAnimation.setDuration(9000)
        upAnimation.setStartValue(5.0)
        upAnimation.setEndValue(45.0)

        downAnimation = QPropertyAnimation(self.m_graph.scene().activeCamera(),
                b'yRotation')
        downAnimation.setDuration(9000)
        downAnimation.setStartValue(45.0)
        downAnimation.setEndValue(5.0)

        self.m_animationCameraY = QSequentialAnimationGroup()
        self.m_animationCameraY.setLoopCount(-1)
        self.m_animationCameraY.addAnimation(upAnimation)
        self.m_animationCameraY.addAnimation(downAnimation)

        self.m_animationCameraX.start()
        self.m_animationCameraY.start()

        self.m_graph.setActiveInputHandler(self.m_inputHandler)

        self.m_selectionTimer = QTimer()
        self.m_selectionTimer.setInterval(10)
        self.m_selectionTimer.timeout.connect(self.triggerSelection)
        self.m_selectionTimer.start()

    def start(self):
        self.addData()

    def addData(self):
        dataArray = []

        stream = QTextStream()
        dataFile = QFile(QFileInfo(__file__).absolutePath() + '/data.txt')
        if dataFile.open(QIODevice.OpenModeFlag.ReadOnly | QIODevice.OpenModeFlag.Text):
            stream.setDevice(dataFile)
            while not stream.atEnd():
                line = stream.readLine()
                if line.startswith('#'):
                    continue

                strList = line.split(',')
                # Each line has three data items: xPos, yPos and zPos values.
                if len(strList) < 3:
                    continue

                position = QVector3D(float(strList[0]), float(strList[1]),
                        float(strList[2]))
                dataArray.append(QScatterDataItem(position))

        self.m_graph.seriesList()[0].dataProxy().resetArray(dataArray)

    def toggleCameraAnimation(self):
        if self.m_animationCameraX.state() != QAbstractAnimation.State.Paused:
            self.m_animationCameraX.pause()
            self.m_animationCameraY.pause()
        else:
            self.m_animationCameraX.resume()
            self.m_animationCameraY.resume()

    def triggerSelection(self):
        self.m_graph.scene().setSelectionQueryPosition(
                self.m_inputHandler.inputPosition())

    def shadowQualityUpdatedByVisual(self, sq):
        self.shadowQualityChanged.emit(int(sq))

    def changeShadowQuality(self, quality):
        self.m_graph.setShadowQuality(QAbstract3DGraph.ShadowQuality(quality))


if __name__ == '__main__':

    import sys

    app = QApplication(sys.argv)
    graph = Q3DScatter()
    container = QWidget.createWindowContainer(graph)

    screenSize = graph.screen().size()
    container.setMinimumSize(
            QSize(int(screenSize.width() / 2), int(screenSize.height() / 1.5)))
    container.setMaximumSize(screenSize)
    container.setSizePolicy(QSizePolicy.Policy.Expanding,
            QSizePolicy.Policy.Expanding)
    container.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

    widget = QWidget()
    hLayout = QHBoxLayout(widget)
    vLayout = QVBoxLayout()
    hLayout.addWidget(container, 1)
    hLayout.addLayout(vLayout)

    widget.setWindowTitle("Custom Input Handling")

    cameraButton = QPushButton("Toggle camera animation")

    shadowQuality = QComboBox()
    shadowQuality.addItem("None")
    shadowQuality.addItem("Low")
    shadowQuality.addItem("Medium")
    shadowQuality.addItem("High")
    shadowQuality.addItem("Low Soft")
    shadowQuality.addItem("Medium Soft")
    shadowQuality.addItem("High Soft")
    shadowQuality.setCurrentIndex(2)

    vLayout.addWidget(cameraButton, 0, Qt.AlignmentFlag.AlignTop)
    vLayout.addWidget(QLabel("Adjust shadow quality"), 0,
            Qt.AlignmentFlag.AlignTop)
    vLayout.addWidget(shadowQuality, 1, Qt.AlignmentFlag.AlignTop)

    modifier = ScatterDataModifier(graph)

    cameraButton.clicked.connect(modifier.toggleCameraAnimation)
    shadowQuality.currentIndexChanged.connect(modifier.changeShadowQuality)
    modifier.shadowQualityChanged.connect(shadowQuality.setCurrentIndex)

    widget.show()
    modifier.start()
    sys.exit(app.exec())
